import logging
import zoneinfo
from dataclasses import replace
from datetime import datetime
from typing import FrozenSet, Iterable, List, Optional, Set, Tuple, Union

from asn1crypto import x509
from pyhanko.sign.ades import qualified_asn1
from pyhanko.sign.ades.report import AdESIndeterminate
from pyhanko.sign.validation.errors import SignatureValidationError
from pyhanko.sign.validation.policy_decl import QualificationRequirements
from pyhanko.sign.validation.qualified.q_status import (
    QcPrivateKeyManagementType,
    QualificationResult,
    QualifiedStatus,
)
from pyhanko.sign.validation.qualified.tsp import (
    CA_QC_URI,
    QTST_URI,
    BaseServiceInformation,
    QcCertType,
    Qualification,
    QualifiedServiceInformation,
    Qualifier,
    TSPRegistry,
)
from pyhanko_certvalidator.authority import TrustedServiceType
from pyhanko_certvalidator.path import ValidationPath

__all__ = [
    'QualificationAssessor',
    'QualificationPolicyError',
    'enforce_requirements',
]


logger = logging.getLogger(__name__)


EIDAS_START_DATE = datetime(
    2016, 7, 1, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo('CET')
)

PRE_EIDAS_QCP_POLICY = '0.4.0.1456.1.1'
PRE_EIDAS_QCP_PLUS_POLICY = '0.4.0.1456.1.2'

UNQUALIFIED = QualifiedStatus(
    qualified=False,
    qc_type=QcCertType.QC_ESIGN,
    qc_key_security=QcPrivateKeyManagementType.UNKNOWN,
)


class QualificationPolicyError(SignatureValidationError):
    """Error triggered by a qualification policy violation."""

    pass


class QualificationAssessor:
    """
    Assesses the qualification status of certificates against a particular
    :class:`.TSPRegistry`.
    """

    def __init__(self, tsp_registry: TSPRegistry):
        self._registry = tsp_registry

    @staticmethod
    def _process_qc_statements(cert: x509.Certificate) -> QualifiedStatus:
        qcs = qualified_asn1.get_qc_statements(cert)
        qualified = False
        key_secure = False
        qc_type = QcCertType.QC_ESIGN
        for statement in qcs:
            st_type = statement['statement_id'].native
            if st_type == 'qc_compliance':
                qualified = True
            elif st_type == 'qc_sscd':
                # management delegation is not encoded by the QcStatements
                key_secure = True
            elif st_type == 'qc_type':
                qc_types: qualified_asn1.QcCertificateType = statement[
                    'statement_info'
                ]
                if len(qc_types) != 1:
                    # In theory this is not limited to one value, we have to
                    # let the TL override in a case like this.
                    # Nonetheless there's really no good reason to do this,
                    # and some ETSI specs are more strict than others,
                    # so I'll deal with this case when it presents itself
                    raise NotImplementedError("only support exactly 1 qc_type")
                qc_type = QcCertType(qc_types[0].native)
        return QualifiedStatus(
            qualified=qualified,
            qc_type=qc_type,
            qc_key_security=(
                QcPrivateKeyManagementType.QSCD
                if key_secure and qualified
                else QcPrivateKeyManagementType.UNKNOWN
            ),
        )

    @staticmethod
    def _check_cd_applicable(
        sd: BaseServiceInformation, putative_status: QualifiedStatus
    ):
        sd_declared_type = sd.additional_info_certificate_type
        if sd_declared_type and putative_status.qc_type not in sd_declared_type:
            logger.info(
                f"Found matching SDI {sd.service_name} on path; "
                f"skipping because QC type does not match"
            )
            return False
        return True

    @staticmethod
    def _apply_sd_qualifications(
        cert: x509.Certificate,
        prelim_status: QualifiedStatus,
        qualifications: Iterable[Qualification],
    ):
        applicable_qualifiers: Set[Qualifier] = set()
        for qualification in qualifications:
            if not qualification.criteria_list.matches(cert):
                continue
            applicable_qualifiers.update(qualification.qualifiers)
        return QualificationAssessor._final_status(
            prelim_status, frozenset(applicable_qualifiers)
        )

    @staticmethod
    def _final_status(
        prelim_status: QualifiedStatus,
        applicable_qualifiers: FrozenSet[Qualifier],
    ):
        # TODO explicitly check consistency / contradictory qualifiers
        # (for now we just use conservative defaults)
        is_qualified: bool
        if (
            Qualifier.NOT_QUALIFIED in applicable_qualifiers
            or Qualifier.LEGAL_PERSON in applicable_qualifiers
        ):
            is_qualified = False
        elif Qualifier.QC_STATEMENT in applicable_qualifiers:
            is_qualified = True
        else:
            is_qualified = prelim_status.qualified

        qc_type: QcCertType
        if Qualifier.FOR_WSA in applicable_qualifiers:
            qc_type = QcCertType.QC_WEB
        elif Qualifier.FOR_ESIG in applicable_qualifiers:
            qc_type = QcCertType.QC_ESIGN
        elif Qualifier.FOR_ESEAL in applicable_qualifiers:
            qc_type = QcCertType.QC_ESEAL
        else:
            qc_type = prelim_status.qc_type

        key_mgmt: QcPrivateKeyManagementType
        if not is_qualified:
            key_mgmt = QcPrivateKeyManagementType.UNKNOWN
        elif (
            Qualifier.NO_SSCD in applicable_qualifiers
            or Qualifier.NO_QSCD in applicable_qualifiers
        ):
            key_mgmt = QcPrivateKeyManagementType.UNKNOWN
        elif Qualifier.QSCD_MANAGED_ON_BEHALF in applicable_qualifiers:
            key_mgmt = QcPrivateKeyManagementType.QSCD_DELEGATED
        elif (
            Qualifier.WITH_SSCD in applicable_qualifiers
            or Qualifier.WITH_QSCD in applicable_qualifiers
        ):
            key_mgmt = QcPrivateKeyManagementType.QSCD
        else:
            key_mgmt = prelim_status.qc_key_security
        return QualifiedStatus(
            qualified=is_qualified,
            qc_type=qc_type,
            qc_key_security=key_mgmt,
        )

    def check_entity_cert_qualified(
        self, path: ValidationPath, moment: Optional[datetime] = None
    ) -> QualificationResult:
        """
        Evaluate the qualified status of a certificate (given a specific
        validation path).

        :param path:
            Validation path to scrutinise.
        :param moment:
            Evaluate the status against the service definitions valid
            at the time specified. If ``None``, take the latest
            available service definitions.
        :return:
            A :class:`.QualificationResult` instance.
        """
        cert = path.leaf
        if not isinstance(cert, x509.Certificate):
            raise NotImplementedError(
                "Only public-key certs are in scope for qualification"
            )
        prelim_status = QualificationAssessor._process_qc_statements(cert)
        path_policies = path.qualified_policies()
        reference_time = moment or datetime.now(tz=zoneinfo.ZoneInfo('CET'))
        if reference_time < EIDAS_START_DATE and path_policies:
            # check QCP / QCP+ policy
            policy_oids = {q.user_domain_policy_id for q in path_policies}
            if PRE_EIDAS_QCP_PLUS_POLICY in policy_oids:
                prelim_status = replace(
                    prelim_status,
                    qualified=True,
                    qc_key_security=(
                        QcPrivateKeyManagementType.QSCD_BY_POLICY
                        if not prelim_status.qc_key_security.is_qscd
                        else prelim_status.qc_key_security
                    ),
                )
            elif PRE_EIDAS_QCP_POLICY in policy_oids:
                prelim_status = replace(prelim_status, qualified=True)

        statuses_found: List[
            Tuple[QualifiedServiceInformation, QualifiedStatus]
        ] = []
        for sd in self._registry.applicable_tsps_on_path(path, reference_time):
            # For this subtlety, see the hanging para in the beginning of
            # section 4 in the CEF eSignature DSS validation algorithm doc
            putative_status = QualificationAssessor._apply_sd_qualifications(
                cert, prelim_status, sd.qualifications
            )
            if QualificationAssessor._check_cd_applicable(
                sd.base_info, putative_status
            ):
                statuses_found.append((sd, putative_status))

        uniq_statuses = set(st for _, st in statuses_found)
        if len(statuses_found) == 1:
            # happy path
            return QualificationResult(
                statuses_found[0][1], service_definition=statuses_found[0][0]
            )
        elif len(uniq_statuses) == 1:
            # TODO gather these warnings somewhere so they can be added
            #  to the validation report
            service_info = ', '.join(
                sd.base_info.service_name for sd, _ in statuses_found
            )
            logger.warning(
                f"Qualification algorithm for {cert.subject.human_friendly} "
                f"reached a consistent conclusion, but through several "
                f"different service definitions: {service_info}"
            )
            return QualificationResult(
                statuses_found[0][1], service_definition=statuses_found[0][0]
            )
        elif not uniq_statuses:
            return QualificationResult(UNQUALIFIED, service_definition=None)
        else:
            service_info = ', '.join(
                sd.base_info.service_name for sd, _ in statuses_found
            )
            logger.warning(
                f"Qualification algorithm for {cert.subject.human_friendly} "
                f"reached contradictory conclusions: {uniq_statuses}. "
                f"Several service definitions were found applicable: "
                f"{service_info}. This certificate will not be considered "
                f"qualified."
            )
            return QualificationResult(UNQUALIFIED, service_definition=None)


def enforce_requirements(
    requirements: QualificationRequirements,
    qualification_result: QualificationResult,
    path: ValidationPath,
):
    """
    Internal method to enforce the requirements of a qualification policy
    during validation.
    """

    cert = path.leaf
    if not isinstance(cert, x509.Certificate):
        raise TypeError("Qualification only makes sense for public-key certs")

    status = qualification_result.status

    err_strs = []
    if not status.qualified:
        raise QualificationPolicyError(
            f"Certificate for {cert.subject.human_friendly} is not qualified",
            ades_subindication=AdESIndeterminate.SIG_CONSTRAINTS_FAILURE,
        )
    assert qualification_result.service_definition is not None
    if requirements.require_service_type is not None:
        service_type_uri = (
            qualification_result.service_definition.base_info.service_type
        )
        service_type: Union[str, TrustedServiceType]
        if isinstance(requirements.require_service_type, str):
            service_type = service_type_uri
        else:
            service_type = {
                CA_QC_URI: TrustedServiceType.CERTIFICATE_AUTHORITY,
                QTST_URI: TrustedServiceType.TIME_STAMPING_AUTHORITY,
            }.get(service_type_uri, TrustedServiceType.UNSUPPORTED)
        if (
            service_type != requirements.require_service_type
            or path.pkix_len > 0
        ):
            err_strs.append(
                f"Certificate {cert.subject.human_friendly} "
                f"is not directly trusted as a service of type "
                f"{requirements.require_service_type}."
            )
    if (
        requirements.permit_cert_types is not None
        and status.qc_type not in requirements.permit_cert_types
    ):
        err_strs.append(
            f"Certificate for {cert.subject.human_friendly} is qualified,"
            f"but the type {status.qc_type.name} is not permitted "
            f"by the requirements. Must be one of "
            f"{', '.join(t.name for t in requirements.permit_cert_types)}."
        )
    if (
        requirements.permit_key_mgmt_types is not None
        and status.qc_key_security not in requirements.permit_key_mgmt_types
    ):
        err_strs.append(
            f"Certificate for {cert.subject.human_friendly} is qualified,"
            f"but the key management type {status.qc_key_security.name} "
            f"is not permitted by the requirements. Must be one of "
            f"{', '.join(t.name for t in requirements.permit_key_mgmt_types)}."
        )

    if err_strs:
        raise QualificationPolicyError(
            "; ".join(err_strs),
            ades_subindication=AdESIndeterminate.SIG_CONSTRAINTS_FAILURE,
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

import tzlocal
from pyhanko.pdf_utils import layout
from pyhanko.pdf_utils.layout import LayoutError
from pyhanko.pdf_utils.text import DEFAULT_BOX_LAYOUT, TextBox, TextBoxStyle
from pyhanko.pdf_utils.writer import BasePdfFileWriter

from .base import BaseStamp, BaseStampStyle

__all__ = ['TextStamp', 'TextStampStyle']


@dataclass(frozen=True)
class TextStampStyle(BaseStampStyle):
    """
    Style for text-based stamps.

    Roughly speaking, this stamp type renders some predefined (but parametrised)
    piece of text inside a text box, and possibly applies a background to it.
    """

    text_box_style: TextBoxStyle = TextBoxStyle()
    """
    The text box style for the internal text box used.
    """

    inner_content_layout: Optional[layout.SimpleBoxLayoutRule] = None
    """
    Rule determining the position and alignment of the inner text box within
    the stamp.

    .. warning::
        This only affects the position of the box, not the alignment of the
        text within.
    """

    stamp_text: str = '%(ts)s'
    """
    Text template for the stamp. The template can contain an interpolation
    parameter ``ts`` that will be replaced by the stamping time.

    Additional parameters may be added if necessary. Values for these must be
    passed to the :meth:`~.TextStamp.__init__` method of the
    :class:`.TextStamp` class in the ``text_params`` argument.
    """

    timestamp_format: str = '%Y-%m-%d %H:%M:%S %Z'
    """
    Datetime format used to render the timestamp.
    """

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'TextStamp':
        return TextStamp(
            writer=writer, style=self, box=box, text_params=text_params
        )


class TextStamp(BaseStamp):
    """
    Class that renders a text stamp as specified by an instance
    of :class:`.TextStampStyle`.
    """

    def __init__(
        self,
        writer: BasePdfFileWriter,
        style,
        text_params=None,
        box: Optional[layout.BoxConstraints] = None,
    ):
        super().__init__(box=box, style=style, writer=writer)
        self.text_params = text_params

        self.text_box: Optional[TextBox] = None

    def get_default_text_params(self):
        """
        Compute values for the default string interpolation parameters
        to be applied to the template string specified in the stamp
        style. This method does not take into account the ``text_params``
        init parameter yet.

        :return:
            A dictionary containing the parameters and their values.
        """
        ts = datetime.now(tz=tzlocal.get_localzone())
        return {
            'ts': ts.strftime(self.style.timestamp_format),
        }

    def _text_layout(self):
        # Set the contents of the text box
        self.text_box = tb = TextBox(
            self.style.text_box_style,
            writer=self.writer,
            resources=self.resources,
            box=None,
        )
        _text_params = self.get_default_text_params()
        if self.text_params is not None:
            _text_params.update(self.text_params)
        try:
            text = self.style.stamp_text % _text_params
        except KeyError as e:
            raise LayoutError(f"Stamp text parameter '{e.args[0]}' is missing")
        tb.content = text

        # Render the text box in its natural size, we'll deal with
        # the minutiae later
        return tb.render()

    def _inner_layout_natural_size(self):
        # render text
        text_commands = self._text_layout()

        inn_box = self.text_box.box
        return [text_commands], (inn_box.width, inn_box.height)

    def _inner_content_layout_rule(self):
        return self.style.inner_content_layout or DEFAULT_BOX_LAYOUT

    def _render_inner_content(self):
        command_stream = [b'q']

        # compute the inner bounding box
        (
            inn_commands,
            (
                inn_width,
                inn_height,
            ),
        ) = self._inner_layout_natural_size()

        inner_layout = self._inner_content_layout_rule()

        bbox = self.box

        # position the inner box
        inn_position = inner_layout.fit(bbox, inn_width, inn_height)

        command_stream.append(inn_position.as_cm())
        command_stream.extend(inn_commands)
        command_stream.append(b'Q')

        return command_stream



import itertools
import logging
import sys
from dataclasses import dataclass
from datetime import datetime
from importlib import resources
from typing import (
    Any,
    FrozenSet,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from asn1crypto import x509
from cryptography.x509 import load_der_x509_certificate
from lxml import etree
from lxml.etree import QName
from pyhanko.generated.etsi import (
    MimeType,
    OtherTSLPointer,
    SchemeTerritory,
    SchemeTypeCommunityRules,
    ServiceStatus,
    ServiceTypeIdentifier,
    TrustServiceStatusList,
    ts_119612,
    ts_119612_extra,
    ts_119612_sie,
    xades,
)
from pyhanko.keys import load_certs_from_pemder_data
from pyhanko.sign.ades.report import AdESFailure, AdESIndeterminate
from pyhanko.sign.validation.errors import SignatureValidationError
from pyhanko.sign.validation.qualified.tsp import (
    _SVCINFOEXT_URI_BASE,
    _TRUSTEDLIST_URI_BASE,
    CA_QC_URI,
    QTST_URI,
    AdditionalServiceInformation,
    BaseServiceInformation,
    CAServiceInformation,
    CertSubjectDNCriterion,
    CriteriaCombination,
    CriteriaList,
    Criterion,
    KeyUsageCriterion,
    PolicySetCriterion,
    QcCertType,
    QTSTServiceInformation,
    Qualification,
    QualifiedServiceInformation,
    Qualifier,
    TSPRegistry,
    TSPServiceParsingError,
)
from pyhanko.sign.validation.settings import KeyUsageConstraints
from signxml.exceptions import SignXMLException as InvalidXmlSignature
from signxml.xades import XAdESSignatureConfiguration, XAdESVerifier
from xsdata.exceptions import XmlContextError
from xsdata.formats.dataclass.models.builders import XmlMetaBuilder
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.handlers import LxmlEventHandler
from xsdata.formats.dataclass.parsers.handlers.lxml import EVENTS

__all__ = [
    'latest_known_lotl_tlso_certs',
    'ojeu_bootstrap_lotl_tlso_certs',
    'parse_lotl_unsafe',
    'read_qualified_service_definitions',
    'trust_list_to_registry',
    'trust_list_to_registry_unsafe',
    'validate_and_parse_lotl',
]

logger = logging.getLogger(__name__)
STATUS_GRANTED = f'{_TRUSTEDLIST_URI_BASE}/Svcstatus/granted'
_SCHEME_RULES_URI_BASE = f'{_TRUSTEDLIST_URI_BASE}/schemerules'
LOTL_RULE = f'{_SCHEME_RULES_URI_BASE}/EUlistofthelists'
ETSI_TSL_MIME_TYPE = 'application/vnd.etsi.tsl+xml'
_CERTIFICATE_TYPE_BY_URI = {
    f'{_SVCINFOEXT_URI_BASE}/ForeSignatures': QcCertType.QC_ESIGN,
    f'{_SVCINFOEXT_URI_BASE}/ForeSeals': QcCertType.QC_ESEAL,
    f'{_SVCINFOEXT_URI_BASE}/ForWebSiteAuthentication': QcCertType.QC_WEB,
}
_QUALIFIER_BY_URI = {q.uri: q for q in Qualifier}
PREFERRED_LANGUAGE: str = 'en'

if sys.version_info >= (3, 14):
    # monkeypatch until there is an xsdata release with this fix:
    # https://github.com/tefra/xsdata/pull/1173

    def _find_declared_class(
        _cls: type, clazz: type, name: str
    ) -> Any:  # pragma: nocover
        import inspect

        for base in clazz.__mro__:
            ann = inspect.get_annotations(base)
            if ann and name in ann:
                return base

        raise XmlContextError(
            f"Failed to detect the declared class for field {name}"
        )

    XmlMetaBuilder.find_declared_class = _find_declared_class  # type: ignore


def _service_name_from_intl_string(
    intl_string: Optional[ts_119612.InternationalNamesType],
) -> str:
    return (
        _extract_from_intl_string(intl_string.name)
        if intl_string
        else "unknown"
    )


def _extract_from_intl_string(
    intl_string: Tuple[
        Union[ts_119612.MultiLangStringType, ts_119612.MultiLangNormStringType],
        ...,
    ],
):
    first_value = intl_string[0].value
    for part in intl_string:
        if part.lang == PREFERRED_LANGUAGE:
            return part.value
    return first_value


T = TypeVar('T')


def _required(thing: Optional[T], descr: str) -> T:
    if thing is None:
        raise TSPServiceParsingError(f"{descr} must be provided")
    return thing


def _as_certs(sdi: ts_119612.ServiceDigitalIdentity):
    for digital_id in sdi.digital_id:
        cert_bytes = digital_id.x509_certificate
        if cert_bytes:
            yield x509.Certificate.load(cert_bytes)


def _process_criteria_list(
    criteria: Optional[ts_119612_sie.CriteriaListType],
) -> CriteriaList:
    entries = frozenset(_process_criteria_list_entries(criteria))
    if entries:
        assert criteria is not None
        assertion_type = _required(
            criteria.assert_value, "Criteria assertion type"
        ).value
        combine_as = CriteriaCombination(str(assertion_type))
        return CriteriaList(combine_as=combine_as, criteria=entries)
    else:
        raise TSPServiceParsingError("No criteria")


def _parse_xades_oid(oid: xades.ObjectIdentifierType) -> str:
    return _required(oid.identifier, "Identifier tag in OID").value


def _process_criteria_list_entries(
    criteria_list: Optional[ts_119612_sie.CriteriaListType],
) -> Generator[Criterion, None, None]:
    if not criteria_list:
        return
    if criteria_list.policy_set:
        # TODO also take policy qualifiers into account
        for policy_set_criterion in criteria_list.policy_set:
            yield PolicySetCriterion(
                required_policy_oids=frozenset(
                    _parse_xades_oid(policy)
                    for policy in policy_set_criterion.policy_identifier
                )
            )
    if criteria_list.key_usage:
        for ku_criterion in criteria_list.key_usage:
            key_usage_must_have = frozenset(
                _required(bit.name, "Key usage bit type name").name.lower()
                for bit in ku_criterion.key_usage_bit
                if bit.value is True
            )
            key_usage_forbidden = frozenset(
                _required(bit.name, "Key usage bit type name").name.lower()
                for bit in ku_criterion.key_usage_bit
                if bit.value is False
            )
            yield KeyUsageCriterion(
                settings=KeyUsageConstraints(
                    key_usage=key_usage_must_have,
                    key_usage_forbidden=key_usage_forbidden,
                )
            )
    sublists: Iterable[ts_119612_sie.CriteriaListType] = (
        criteria_list.criteria_list
    )
    if sublists:
        for sublist in sublists:
            yield _process_criteria_list(sublist)
    if criteria_list.other_criteria_list:
        for el in criteria_list.other_criteria_list.content:
            if isinstance(el, ts_119612_extra.CertSubjectDNAttribute):
                yield CertSubjectDNCriterion(
                    frozenset(
                        _parse_xades_oid(oid_xml)
                        for oid_xml in el.attribute_oid
                    )
                )
            elif isinstance(el, ts_119612_extra.ExtendedKeyUsage):
                kpids = frozenset(
                    _parse_xades_oid(kpid_xml) for kpid_xml in el.key_purpose_id
                )
                yield KeyUsageCriterion(
                    settings=KeyUsageConstraints(extd_key_usage=kpids)
                )
            else:
                raise TSPServiceParsingError(
                    f"Unknown criterion {el} in qualifier definition"
                )


def _process_qualifiers(qualifiers: Iterable[ts_119612_sie.QualifierType]):
    for qual in qualifiers:
        try:
            yield _QUALIFIER_BY_URI[qual.uri]
        except KeyError:
            logger.info(f"Qualifier {qual.uri} in SDI ignored...")


def _get_qualifications(qualifications: ts_119612_sie.Qualifications):
    for qual in qualifications.qualification_element:
        quals_xml = qual.qualifiers
        if not quals_xml:
            continue
        # note: all the qualifiers we currently support only make sense
        # for qualified CAs, not other types of services
        # (qualified or otherwise).
        qualifiers = frozenset(_process_qualifiers(quals_xml.qualifier))
        criteria = _process_criteria_list(qual.criteria_list)
        if qualifiers and criteria:
            yield Qualification(qualifiers=qualifiers, criteria_list=criteria)
        else:
            logger.warning(f"Could not process qualifiers in {quals_xml}")


def _process_additional_info(
    info: ts_119612.AdditionalServiceInformation, critical: bool
):
    return AdditionalServiceInformation(
        uri=_required(info.uri, "Additional service info URI").value,
        textual_info=info.information_value,
        critical=critical,
    )


def _interpret_service_info_for_ca(
    service: ts_119612.TSPService,
):
    return _interpret_service_info_for_tsp(
        service, _interpret_historical_service_info_for_ca
    )


def _interpret_service_info_for_tsp(
    service: ts_119612.TSPService,
    historical_interpreter,
):
    service_info = service.service_information
    assert service_info is not None
    return historical_interpreter(
        service_info=ts_119612.ServiceHistoryInstance(
            service_type_identifier=service_info.service_type_identifier,
            service_name=service_info.service_name,
            service_digital_identity=service_info.service_digital_identity,
            service_status=service_info.service_status,
            status_starting_time=service_info.status_starting_time,
            service_information_extensions=service_info.service_information_extensions,
        ),
        next_update_at=None,
    )


def _interpret_historical_service_info_for_ca(
    service_info: ts_119612.ServiceHistoryInstance,
    next_update_at: Optional[datetime],
):
    certs = list(
        _as_certs(
            _required(
                service_info.service_digital_identity,
                "Service digital identity",
            )
        )
    )
    service_name = _service_name_from_intl_string(service_info.service_name)
    qualifications: FrozenSet[Qualification] = frozenset()
    expired_revinfo_date = None
    additional_info = []
    extensions_xml = (
        service_info.service_information_extensions.extension
        if service_info.service_information_extensions
        else ()
    )
    asi_qc_type: Set[QcCertType] = set()
    for ext in extensions_xml:
        for ext_content in ext.content:
            if isinstance(ext_content, ts_119612_sie.Qualifications):
                qualifications = frozenset(_get_qualifications(ext_content))
            elif isinstance(ext_content, ts_119612.ExpiredCertsRevocationInfo):
                expired_revinfo_date = _required(
                    ext_content.value,
                    "Date in expired certs revocation info cutoff",
                ).to_datetime()
            elif isinstance(
                ext_content, ts_119612.AdditionalServiceInformation
            ):
                additional_info_entry = _process_additional_info(
                    ext_content, ext.critical or False
                )
                try:
                    asi_qc_type.add(
                        _CERTIFICATE_TYPE_BY_URI[additional_info_entry.uri]
                    )
                except KeyError:
                    additional_info.append(additional_info_entry)
            elif ext.critical:
                raise TSPServiceParsingError(
                    f"Cannot process a critical extension "
                    f"in service named '{service_name}'.\n"
                    f"Content: {ext_content}"
                )
    valid_from_date = service_info.status_starting_time
    if valid_from_date is None:
        raise TSPServiceParsingError(
            f"The validity start of the status of "
            f"the the service named {service_name} is not known. "
            f"This is an error."
        )
    base_service_info = BaseServiceInformation(
        service_type=_required(
            service_info.service_type_identifier, "Service type identifier"
        ).value,
        valid_from=valid_from_date.to_datetime(),
        valid_until=next_update_at,
        service_name=service_name,
        provider_certs=tuple(certs),
        additional_info_certificate_type=frozenset(asi_qc_type),
        other_additional_info=frozenset(additional_info),
    )

    return CAServiceInformation(
        base_info=base_service_info,
        qualifications=qualifications,
        expired_certs_revocation_info=expired_revinfo_date,
    )


def _interpret_historical_service_info_for_qtst(
    service_info: ts_119612.ServiceHistoryInstance,
    next_update_at: Optional[datetime],
):
    # todo dedupe
    certs = list(
        _as_certs(
            _required(
                service_info.service_digital_identity,
                "Service digital identity",
            )
        )
    )
    service_name = _service_name_from_intl_string(service_info.service_name)
    qualifications: FrozenSet[Qualification] = frozenset()
    additional_info: List[AdditionalServiceInformation] = []
    extensions_xml = (
        service_info.service_information_extensions.extension
        if service_info.service_information_extensions
        else ()
    )
    for ext in extensions_xml:
        for ext_content in ext.content:
            if isinstance(ext_content, ts_119612_sie.Qualifications):
                qualifications = frozenset(_get_qualifications(ext_content))
            elif ext.critical:
                raise TSPServiceParsingError(
                    f"Cannot process a critical extension "
                    f"in service named '{service_name}'.\n"
                    f"Content: {ext_content}"
                )
    valid_from_date = service_info.status_starting_time
    if valid_from_date is None:
        raise TSPServiceParsingError(
            f"The validity start of the status of "
            f"the the service named {service_name} is not known. "
            f"This is an error."
        )
    base_service_info = BaseServiceInformation(
        service_type=_required(
            service_info.service_type_identifier, "Service type identifier"
        ).value,
        valid_from=valid_from_date.to_datetime(),
        valid_until=next_update_at,
        service_name=service_name,
        provider_certs=tuple(certs),
        additional_info_certificate_type=frozenset(),
        other_additional_info=frozenset(additional_info),
    )

    return QTSTServiceInformation(
        base_info=base_service_info,
        qualifications=qualifications,
    )


def _read_service_history(
    history_items, validity_start, service_name, interpreter
):
    errors_encountered = []
    item_index_sorted_by_date = sorted(
        (
            (orig_ix, item.status_starting_time.to_datetime())
            for orig_ix, item in enumerate(history_items)
            if item.status_starting_time
        ),
        key=lambda t: t[1],
        reverse=True,
    )
    end_of_validity_by_orig_ix = {
        orig_ix: next_start
        for (orig_ix, cur_start), next_start in zip(
            item_index_sorted_by_date,
            itertools.chain(
                (validity_start.to_datetime(),),
                (st for _, st in item_index_sorted_by_date[:-1]),
            ),
        )
    }

    for orig_ix, validity_end in end_of_validity_by_orig_ix.items():
        history_item = history_items[orig_ix]
        if history_item.service_status != ServiceStatus(STATUS_GRANTED):
            continue
        try:
            validity_end = end_of_validity_by_orig_ix[orig_ix]
            yield interpreter(
                history_item,
                next_update_at=validity_end,
            )
        except TSPServiceParsingError as e:
            logger.debug(
                f"Failed to parse item {orig_ix + 1} in history "
                f"of service {service_name}. This history "
                f"entry will not be processed further.",
                exc_info=e,
            )
            errors_encountered.append(e)
    return errors_encountered


def _interpret_service_info_for_tsps(
    services: Iterable[ts_119612.TSPService],
    service_type: ServiceTypeIdentifier,
    interpreter,
):
    errors_encountered = []
    for service in services:
        service_info = service.service_information
        if (
            not service_info
            or service_info.service_type_identifier != service_type
        ):
            continue

        service_name = _service_name_from_intl_string(service_info.service_name)
        # TODO allow the user to specify if they also want to include
        #  other statuses (e.g. national level)
        # TODO evaluate historical definitions too in case of point-in-time
        #  work, store that info on the object
        if service_info.service_status == ServiceStatus(STATUS_GRANTED):
            try:
                yield _interpret_service_info_for_tsp(service, interpreter)
            except TSPServiceParsingError as e:
                logger.warning(
                    f"Failed to process current status "
                    f"of service {service_name}. This history "
                    f"entry will not be processed further.",
                    exc_info=e,
                )
                errors_encountered.append(e)
                continue

        validity_start = service_info.status_starting_time
        if validity_start and service.service_history:
            history_items = service.service_history.service_history_instance
            history_errors = yield from _read_service_history(
                history_items, validity_start, service_name, interpreter
            )
            errors_encountered.extend(history_errors)

    return errors_encountered


class _RestrictedLxmlEventHandler(LxmlEventHandler):
    # Disable xinclude support and entity resolution

    def parse(self, source: Any, ns_map: dict[Optional[str], str]) -> Any:
        if isinstance(source, (etree._ElementTree, etree._Element)):
            ctx = etree.iterwalk(source, EVENTS)
        else:
            ctx = etree.iterparse(
                source,
                EVENTS,
                recover=True,
                remove_comments=True,
                load_dtd=self.parser.config.load_dtd,
                resolve_entities=False,
                no_network=True,
            )

        return self.process_context(ctx, ns_map)


def _raw_tl_parse(tl_xml: str) -> ts_119612.TrustServiceStatusList:
    parser = XmlParser(
        config=ParserConfig(
            load_dtd=False,
            process_xinclude=False,
            fail_on_unknown_properties=False,
            fail_on_unknown_attributes=False,
        ),
        handler=_RestrictedLxmlEventHandler,
    )
    return parser.from_string(tl_xml, clazz=ts_119612.TrustServiceStatusList)


def read_qualified_service_definitions(
    tl_xml: str,
) -> Generator[QualifiedServiceInformation, None, List[TSPServiceParsingError]]:
    """
    Internal function to read qualified service definitions from a trusted list
    payload (does not include signature validation).
    """

    parse_result = _raw_tl_parse(tl_xml)
    tspl = parse_result.trust_service_provider_list
    errors_encountered = []
    for tsp in _required(tspl, "TSP list").trust_service_provider:
        tsp_errors = yield from _interpret_service_info_for_tsps(
            _required(tsp.tspservices, "TSP services (CA)").tspservice,
            service_type=ServiceTypeIdentifier(CA_QC_URI),
            interpreter=_interpret_historical_service_info_for_ca,
        )
        errors_encountered.extend(tsp_errors)

        tsp_errors = yield from _interpret_service_info_for_tsps(
            _required(tsp.tspservices, "TSP services (QTST)").tspservice,
            service_type=ServiceTypeIdentifier(QTST_URI),
            interpreter=_interpret_historical_service_info_for_qtst,
        )
        errors_encountered.extend(tsp_errors)
    return errors_encountered


def trust_list_to_registry_unsafe(
    tl_xml: str, registry: Optional[TSPRegistry] = None
) -> Tuple[TSPRegistry, List[TSPServiceParsingError]]:
    """
    Parse a trusted list (ETSI TS 119 612) into a :class:`TSPRegistry`.

    .. warning::
        The XML signature on the trusted list is _not_ validated as part
        of this method. Currently, ensuring the integrity of the trusted list
        is the caller's responsibility.

    :param tl_xml:
        XML payload describing the trusted list in the ETSI TS 119 612 format
    :param registry:
        Registry to which the entries should be added. If not supplied, a new
        one will be created.
    :return:
    """
    registry = registry or TSPRegistry()
    g = read_qualified_service_definitions(tl_xml)
    while True:
        try:
            sd = next(g)
            if isinstance(sd, CAServiceInformation):
                registry.register_ca(sd)
            elif isinstance(sd, QTSTServiceInformation):
                registry.register_tst(sd)
        except StopIteration as e:
            return registry, e.value


def _verify_xml(tl_xml: str, tlso_cert: x509.Certificate):
    tl_xml_bytes = tl_xml.encode('utf8')
    verifier = XAdESVerifier()
    cert_obj = load_der_x509_certificate(tlso_cert.dump())
    config = XAdESSignatureConfiguration(
        require_x509=True,
        expect_references=True,
    )
    try:
        verify_results = verifier.verify(
            tl_xml_bytes,
            x509_cert=cert_obj,
            expect_config=config,
        )
    except InvalidXmlSignature as e:
        raise SignatureValidationError(
            f"Invalid XML signature on trusted list: {e}",
            ades_subindication=AdESIndeterminate.GENERIC,
        ) from e
    return verify_results


def _validate_and_extract_tl_data(
    tl_xml: str, tlso_cert: x509.Certificate
) -> str:
    verify_results = _verify_xml(tl_xml, tlso_cert)
    tl_signed_xml = None
    for result in verify_results:
        if result.signed_xml is None:
            continue
        if result.signed_xml.tag == QName(
            TrustServiceStatusList.Meta.namespace, 'TrustServiceStatusList'
        ):
            tl_signed_xml = result.signed_data.decode("utf-8")
            break
    if not tl_signed_xml:
        raise SignatureValidationError(
            "Failed to identify trusted list in signed data",
            ades_subindication=AdESFailure.FORMAT_FAILURE,
        )

    logger.debug(
        "Validated TL data with TLSO certificate %s\n(SHA-256 fingerprint %s)",
        tlso_cert.subject.human_friendly,
        tlso_cert.sha256_fingerprint,
    )
    return tl_signed_xml


def _validate_and_extract_tl_data_multiple_certs(
    tl_xml: str, tlso_cert_candidates: List[x509.Certificate]
) -> str:
    # sort the issuer certs by newest first
    sorted_candidates = sorted(
        tlso_cert_candidates, key=lambda c: c.not_valid_before, reverse=True
    )
    errors = []
    result = None
    for candidate in sorted_candidates:
        try:
            result = _validate_and_extract_tl_data(tl_xml, candidate)
            break
        except SignatureValidationError as e:
            errors.append(e)
    if not result:
        msg = (
            f"None of the candidate TLSO certs could be used to validate "
            f"the TL signature: {','.join(e.failure_message for e in errors)}"
        )
        raise SignatureValidationError(
            msg,
            ades_subindication=AdESIndeterminate.GENERIC,
        )
    return result


def trust_list_to_registry(
    tl_xml: str,
    tlso_certs: List[x509.Certificate],
    registry: Optional[TSPRegistry] = None,
) -> Tuple[TSPRegistry, List[TSPServiceParsingError]]:
    """
    Validate and parse a trusted list (ETSI TS 119 612) into a :class:`TSPRegistry`.

    .. note::
        The TSLO certs are used in the validation of the trusted list,
        but are not validated by this function.

    :param tl_xml:
        XML payload describing the trusted list in the ETSI TS 119 612 format
    :param tlso_certs:
        The certificates containing the public keys with which
        the TL could be validated.
    :param registry:
        Registry to which the entries should be added. If not supplied, a new
        one will be created.
    :raises SignatureValidationError:
        If the trusted list's signature cannot be validated.
    :return:
    """
    tl_signed_xml = _validate_and_extract_tl_data_multiple_certs(
        tl_xml, tlso_certs
    )
    return trust_list_to_registry_unsafe(tl_signed_xml, registry)


def _parse_tsl_info(pointer: OtherTSLPointer):
    rules = set()
    territory = list_type = None
    additional_info = _required(
        pointer.additional_information, "TSL pointer additional information"
    )
    for other_info in additional_info.other_information:
        for item in other_info.content:
            if isinstance(item, SchemeTypeCommunityRules):
                for uri in item.uri:
                    rules.add(uri.value)
            elif isinstance(item, SchemeTerritory):
                territory = item.value
            elif isinstance(item, MimeType):
                list_type = item.value
    return pointer.tsllocation, list_type, territory, rules


@dataclass(frozen=True)
class TLReference:
    """
    Reference to a trusted list.
    """

    location_uri: str
    """
    URI where the trusted list can be found.
    """

    territory: str
    """
    Territory with which the references trusted list is assocated.
    """

    tlso_certs: List[x509.Certificate]
    """
    Certificates that can be used to validate the signature
    on the referenced trusted list.
    """

    scheme_rules: FrozenSet[str]
    """
    URIs for scheme rules that apply to the referenced trusted list.
    """


@dataclass(frozen=True)
class LOTLParseResult:
    references: List[TLReference]
    errors: List[TSPServiceParsingError]
    pivot_urls: List[str]


def parse_lotl_unsafe(
    lotl_xml: str,
) -> LOTLParseResult:
    """
    Parse a list-of-the-lists (LOTL).

    .. warning::
        This function does not include validating
        the signature on the LOTL.

    :param lotl_xml:
        XML LOTL payload
    :return:
        A parse result.
    """
    parse_result = _raw_tl_parse(lotl_xml)

    scheme_info = _required(
        parse_result.scheme_information, "LOTL scheme information"
    )
    pointers = _required(
        scheme_info.pointers_to_other_tsl, "pointers to other TSLs"
    )
    info_uris = _required(
        scheme_info.scheme_information_uri, "scheme information URIs"
    )

    references = []
    errors = []
    for pointer in pointers.other_tslpointer:
        location, list_type, territory, rules = _parse_tsl_info(pointer)
        if (
            list_type != ETSI_TSL_MIME_TYPE
            or location is None
            or territory is None
            or pointer.service_digital_identities is None
        ):
            continue
        tl_issuer_certs = []
        for (
            digital_identity
        ) in pointer.service_digital_identities.service_digital_identity:
            for sdi in digital_identity.digital_id:
                cert_bytes = sdi.x509_certificate
                if cert_bytes is not None:
                    try:
                        tl_issuer_certs.append(
                            x509.Certificate.load(cert_bytes)
                        )
                    except Exception as e:
                        errors.append(
                            TSPServiceParsingError(
                                f"Failed to load certificate for {territory} ({location}): {e}"
                            )
                        )
        ref = TLReference(
            location, territory, tl_issuer_certs, frozenset(rules)
        )
        references.append(ref)
    pivots = [
        info_uri.value
        for info_uri in info_uris.uri
        if info_uri.value.endswith(".xml")
    ]
    return LOTLParseResult(
        references=references, errors=errors, pivot_urls=pivots
    )


def _lotl_certs_file(fname: str) -> List[x509.Certificate]:
    pkg_files = resources.files("pyhanko.sign.validation.qualified")
    data = pkg_files.joinpath("lotl-certs", fname).read_bytes()
    return list(load_certs_from_pemder_data(data))


def latest_known_lotl_tlso_certs() -> List[x509.Certificate]:
    """
    Retrieve the lastest known (at the time of the most recent pyHanko release)
    list of LOTL signer certs.
    """
    return _lotl_certs_file("latest.cert.pem")


def ojeu_bootstrap_lotl_tlso_certs() -> List[x509.Certificate]:
    """
    Retrieve the list of certificates published
    in `OJ C 276, 16.8.2019 <https://eur-lex.europa.eu/legal-content/EN/TXT/?uri=uriserv:OJ.C_.2019.276.01.0001.01.ENG>`_,
    which is bundled with this library.
    """
    return _lotl_certs_file("bootstrap.cert.pem")


# TODO check validity time windows


def validate_and_parse_lotl(
    lotl_xml: str, lotl_tlso_certs: Optional[List[x509.Certificate]] = None
) -> LOTLParseResult:
    """
    Validate and parse a list-of-the-lists (LOTL).

    :param lotl_xml:
        XML LOTL payload
    :param lotl_tlso_certs:
        List of certificates that can be used to validate the list-of-the-lists.
        If not specified, the list-of-the-lists will be validated against the
        last known set of list-of-the-lists signer certs bundled with this
        library.

        See :func:`validate_and_parse_lotl`.
    :return:
        A parse result.
    """

    if not lotl_tlso_certs:
        lotl_tlso_certs = latest_known_lotl_tlso_certs()
    lotl_xml_validated = _validate_and_extract_tl_data_multiple_certs(
        lotl_xml, lotl_tlso_certs
    )
    return parse_lotl_unsafe(lotl_xml_validated)



from dataclasses import dataclass, field
from typing import Optional, Tuple

from .ts_119612 import (
    AnyType,
    InternationalNamesType,
    NonEmptyMultiLangURIType,
    SchemeOperatorName,
)
from .xades import ObjectIdentifierType

__NAMESPACE__ = "http://uri.etsi.org/02231/v2/additionaltypes#"


@dataclass(frozen=True)
class MimeType:
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"

    value: str = field(
        default="",
        metadata={
            "required": True,
        },
    )


@dataclass(frozen=True)
class PublicKeyLocation:
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"

    value: str = field(
        default="",
        metadata={
            "required": True,
            "min_length": 1,
        },
    )


@dataclass(frozen=True)
class X509CertificateLocation:
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"

    value: str = field(
        default="",
        metadata={
            "required": True,
            "min_length": 1,
        },
    )


@dataclass(frozen=True)
class CertSubjectDNAttributeType:
    attribute_oid: Tuple[ObjectIdentifierType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "AttributeOID",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2/additionaltypes#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class ExtendedKeyUsageType:
    key_purpose_id: Tuple[ObjectIdentifierType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "KeyPurposeId",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2/additionaltypes#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class TakenOverByType:
    uri: Optional[NonEmptyMultiLangURIType] = field(
        default=None,
        metadata={
            "name": "URI",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2/additionaltypes#",
            "required": True,
        },
    )
    tspname: Optional[InternationalNamesType] = field(
        default=None,
        metadata={
            "name": "TSPName",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2/additionaltypes#",
            "required": True,
        },
    )
    scheme_operator_name: Optional[SchemeOperatorName] = field(
        default=None,
        metadata={
            "name": "SchemeOperatorName",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2#",
            "required": True,
        },
    )
    scheme_territory: Optional[str] = field(
        default=None,
        metadata={
            "name": "SchemeTerritory",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2#",
            "required": True,
        },
    )
    other_qualifier: Tuple[AnyType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "OtherQualifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/02231/v2/additionaltypes#",
        },
    )


@dataclass(frozen=True)
class CertSubjectDNAttribute(CertSubjectDNAttributeType):
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"


@dataclass(frozen=True)
class ExtendedKeyUsage(ExtendedKeyUsageType):
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"


@dataclass(frozen=True)
class TakenOverBy(TakenOverByType):
    class Meta:
        namespace = "http://uri.etsi.org/02231/v2/additionaltypes#"



import asyncio
import hashlib
import json
import logging
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import aiohttp
from aiohttp import ClientTimeout
from asn1crypto import x509
from pyhanko.sign.validation.qualified import eutl_parse
from pyhanko.sign.validation.qualified.eutl_parse import LOTL_RULE
from pyhanko.sign.validation.qualified.tsp import (
    TSPRegistry,
    TSPServiceParsingError,
)

__all__ = [
    'EU_LOTL_LOCATION',
    'FileSystemTLCache',
    'InMemoryTLCache',
    'TLCache',
    'bootstrap_lotl_signers',
    'fetch_lotl',
    'lotl_to_registry',
]

logger = logging.getLogger(__name__)

EU_LOTL_LOCATION = 'https://ec.europa.eu/tools/lotl/eu-lotl.xml'
"""
Location of the EU's global list-of-the-lists (LOTL).

.. note::
    At the time of writing, this URL appears to have remained constant
    since eIDAS's inception, but in theory it is subject to change.
"""


class TLCache:
    """
    Cache for trusted lists, intended to speed up downloading lists
    from a list-of-lists.

    This is internal API.
    """

    def __getitem__(self, key: str) -> str:
        raise NotImplementedError

    def __setitem__(self, key: str, value: str) -> None:
        raise NotImplementedError


class InMemoryTLCache(TLCache):
    """
    Cache for trusted lists, intended to speed up downloading lists
    from a list-of-lists.

    The cache is keyed by download URL and does not have any eviction mechanism.
    """

    def __init__(self: 'InMemoryTLCache'):
        self._cache: Dict[str, str] = {}

    def __getitem__(self, key: str) -> str:
        return self._cache[key]

    def __setitem__(self, key: str, value: str) -> None:
        self._cache[key] = value


class FileSystemTLCache(TLCache):
    def __init__(self, cache_path: Path, expire_after: timedelta):
        self._cache: Dict[str, Tuple[datetime, str]] = {}
        self._root = cache_path
        self._expire_after = expire_after
        if not cache_path.exists():
            cache_path.mkdir(parents=True, exist_ok=True)

        index = cache_path / 'index.json'
        if index.exists():
            with index.open('r') as inf:
                index_data = json.load(inf)
                for key, entry in index_data.items():
                    exp_ts = datetime.fromtimestamp(
                        entry['exp_epoch_seconds'], tz=timezone.utc
                    )
                    self._cache[key] = (exp_ts, entry['fname'])
        logger.debug(
            f"Loaded {len(self._cache)} items from cache at {cache_path.absolute()}"
        )

    def __getitem__(self, key: str) -> str:
        exp_ts, fname = self._cache[key]
        now = datetime.now(timezone.utc)
        if now > exp_ts:
            raise KeyError
        cached_file_path = self._root / fname
        try:
            with cached_file_path.open('r') as inf:
                content = inf.read()
        except IOError as e:
            logger.warning(
                f"Failed to access cached file at {cached_file_path}: {e}",
                exc_info=e,
            )
            raise KeyError
        return content

    def __setitem__(self, key: str, value: str) -> None:
        exp_ts = datetime.now(timezone.utc) + self._expire_after
        fname = hashlib.sha256(key.encode('utf8')).hexdigest()
        index = self._root / 'index.json'
        if index.exists():
            with index.open('r') as inf:
                index_data = json.load(inf)
        else:
            index_data = {}
        index_data[key] = {
            'exp_epoch_seconds': exp_ts.timestamp(),
            'fname': fname,
        }
        with index.open('w') as outf:
            json.dump(index_data, outf)
        with (self._root / fname).open('w') as outf:
            outf.write(value)
        self._cache[key] = (exp_ts, fname)

    def reset(self):
        for fname in self._root.iterdir():
            fname.unlink(missing_ok=True)


FETCH_TRIES = 3
FETCH_BASE_DELAY_SECONDS = 2
FETCH_TIMEOUT_SECONDS = 30
FETCH_CONNECT_TIMEOUT_SECONDS = 2


async def _get(uri: str, client: aiohttp.ClientSession) -> str:
    delay_s = FETCH_BASE_DELAY_SECONDS
    last_error = None
    for attempt in range(FETCH_TRIES):
        try:
            response = await client.get(
                uri,
                headers=(
                    ('Accept', 'text/xml'),
                    ('Accept', eutl_parse.ETSI_TSL_MIME_TYPE),
                ),
                raise_for_status=True,
                timeout=ClientTimeout(
                    total=FETCH_TIMEOUT_SECONDS,
                    sock_read=FETCH_CONNECT_TIMEOUT_SECONDS,
                ),
            )
            return await response.text()
        except aiohttp.ClientError as e:
            if isinstance(e, aiohttp.ClientResponseError) and e.status < 500:
                raise e
            last_error = e
            if attempt < FETCH_TRIES - 1:
                await asyncio.sleep(delay_s)
                delay_s *= 2
    assert last_error is not None
    raise last_error


async def _fetch(
    cache: Optional[TLCache], uri: str, client: aiohttp.ClientSession
):
    if cache is None:
        return await _get(uri, client)
    else:
        try:
            result = cache[uri]
            logger.info(f"Retrieved {uri} from cache")
            return result
        except KeyError:
            pass
        result = await _get(uri, client)
        cache[uri] = result
        return result


async def bootstrap_lotl_signers(
    latest_lotl_xml: str,
    client: aiohttp.ClientSession,
    bootstrap_lotl_tlso_certs: Optional[List[x509.Certificate]] = None,
    cache: Optional[TLCache] = None,
) -> List[x509.Certificate]:
    """
    Perform the bootstrapping process specified in
    Article 4 of Commission Implementing Decision (EU) 2015/1505 to determine
    the certificates that can be used to verify a list-of-the-lists
    signature.

    The EC's `technical guidance <https://ec.europa.eu/tools/lotl/pivot-lotl-explanation.html>`_
    on this topic is also of interest.

    .. warning::
        PyHanko bundles the both the initial set of list-of-the-lists signing
        certificates and the last-known set of such certificates.
        In principle, you only need to use this function if there have not
        been any pyHanko releases since the last change in the list-of-the-lists
        signers, or you cannot upgrade for some reason.

        The process to fetch and process all relevant trust lists is rather
        slow, so this bootstrapping logic should be used judiciously.

    :param latest_lotl_xml:
        The XML content of the most recent list-of-the-lists.
    :param client:
        An :class:`aiohttp.ClientSession` object to use for fetching pivot
        lists-of-the-lists.
    :param bootstrap_lotl_tlso_certs:
        Initial list of certificates. This defaults to the list published
        in `OJ C 276, 16.8.2019 <https://eur-lex.europa.eu/legal-content/EN/TXT/?uri=uriserv:OJ.C_.2019.276.01.0001.01.ENG>`_,
        which is bundled with this library.
    :param cache:
        An optional :class:`TLCache` to be used while fetching pivot
        lists-of-the-lists.
    :return:
        The list of certificates that can be used to verify the most recent
        list-of-the-lists.
    """

    pivots = eutl_parse.parse_lotl_unsafe(latest_lotl_xml).pivot_urls
    sorted_pivots = sorted(pivots)
    if bootstrap_lotl_tlso_certs:
        current_certs = bootstrap_lotl_tlso_certs
    else:
        current_certs = eutl_parse.ojeu_bootstrap_lotl_tlso_certs()

    for pivot in sorted_pivots:
        logger.info(f"Processing LOTL pivot {pivot}...")
        try:
            pivot_xml = await _fetch(cache, pivot, client)
            pivot_parse_result = eutl_parse.validate_and_parse_lotl(
                pivot_xml, current_certs
            )
        except Exception as e:
            raise TSPServiceParsingError(
                f"Failed to process pivot LOTL {pivot}: {e}"
            ) from e

        try:
            lotl_self_reference = next(
                ref
                for ref in pivot_parse_result.references
                if eutl_parse.LOTL_RULE in ref.scheme_rules
            )
        except StopIteration:
            raise TSPServiceParsingError(
                f"Failed to process pivot LOTL {pivot}: could not extract TLSO certs"
            )

        current_certs = lotl_self_reference.tlso_certs
    return current_certs


async def fetch_lotl(
    client: aiohttp.ClientSession,
    cache: Optional[TLCache] = None,
    url=EU_LOTL_LOCATION,
):
    """
    Fetch the EU list-of-the-lists (LOTL).

    :param client:
        An :class:`aiohttp.ClientSession` object to use for fetching
        trust lists.
    :param cache:
        An optional :class:`TLCache` to be used while fetching trust lists.
    :param url:
        The URL content of the list-of-the-lists. The default is
        the location specified in :const:`EU_LOTL_LOCATION`.
    :return:
    """
    logger.info(f"Downloading LOTL from {url}...")
    lotl_xml = await _fetch(cache, url, client)
    return lotl_xml


async def lotl_to_registry(
    lotl_xml: Optional[str],
    client: aiohttp.ClientSession,
    lotl_tlso_certs: Optional[List[x509.Certificate]] = None,
    cache: Optional[TLCache] = None,
    registry: Optional[TSPRegistry] = None,
    only_territories: Optional[Set[str]] = None,
) -> Tuple[TSPRegistry, List[TSPServiceParsingError]]:
    """
    Populate a :class:`.TSPRegistry` instance from a list-of-the-lists XML
    payload, validating the signatures on the trusted lists in the process.

    :param lotl_xml:
        The XML content of a list-of-the-lists. If ``None``, it will
        be downloaded from the location specified in :const:`EU_LOTL_LOCATION`.
    :param client:
        An :class:`aiohttp.ClientSession` object to use for fetching
        trust lists.
    :param lotl_tlso_certs:
        List of certificates that can be used to validate the list-of-the-lists.
        If not specified, the list-of-the-lists will be validated against the
        last known set of list-of-the-lists signer certs bundled with this
        library.

        See :func:`bootstrap_lotl_signers` and
        :func:`~.eutl_parse.validate_and_parse_lotl`.
    :param cache:
        An optional :class:`TLCache` to be used while fetching trust lists.
    :param registry:
        An optional :class:`.TSPRegistry` to be used to collect
        service definitions for trusted service providers.
        If not supplied, a registry will be instantiated.
    :param only_territories:
        Limit downloads to the territories specified (as
        two-letter ISO 3166-1 country codes).
        If ``None``, no filtering is applied.
    :return:
        A :class:`.TSPRegistry` instance populated with the
        fetched contents, in addition to any parsing errors encountered.
    """
    if lotl_xml is None:
        lotl_xml = await fetch_lotl(client, cache)

    lotl_result = eutl_parse.validate_and_parse_lotl(lotl_xml, lotl_tlso_certs)
    errors = lotl_result.errors
    registry = registry or TSPRegistry()
    if only_territories is not None:
        territory_list = {x.casefold() for x in only_territories}
    else:
        territory_list = None
    for ref in lotl_result.references:
        if (
            territory_list is not None
            and ref.territory.casefold() not in territory_list
        ):
            continue
        if LOTL_RULE in ref.scheme_rules:
            continue
        logger.info(
            f"Processing trusted list for {ref.territory} "
            f"at {ref.location_uri}..."
        )
        try:
            tl_xml = await _fetch(cache, ref.location_uri, client)
        except Exception as e:
            errors.append(
                TSPServiceParsingError(
                    f"Failed to download trusted list for {ref.territory} "
                    f"at {ref.location_uri}: {e}"
                )
            )
            continue
        try:
            _, tl_errors = eutl_parse.trust_list_to_registry(
                tl_xml, ref.tlso_certs, registry
            )
            errors.extend(tl_errors)
        except Exception as e:
            errors.append(
                TSPServiceParsingError(
                    f"Failed to parse trusted list for {ref.territory} "
                    f"at {ref.location_uri}: {e}"
                )
            )
    return registry, errors



import uuid
from binascii import hexlify
from dataclasses import dataclass
from typing import Optional, Tuple

from pyhanko.config.api import ConfigurableMixin
from pyhanko.config.errors import ConfigurationError
from pyhanko.pdf_utils import content, generic, layout
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.misc import rd
from pyhanko.pdf_utils.writer import BasePdfFileWriter

from ..pdf_utils.content import AppearanceContent
from .appearances import CoordinateSystem
from .art import STAMP_ART_CONTENT


def _get_background_content(bg_spec) -> content.PdfContent:
    if not isinstance(bg_spec, str):
        raise ConfigurationError("Background specification must be a string")
    # 'special' value to use the stamp vector image baked into
    # the module
    if bg_spec == '__stamp__':
        return STAMP_ART_CONTENT
    elif bg_spec.endswith('.pdf'):
        # import first page of PDF as background
        return content.ImportedPdfPage(bg_spec)
    else:
        from PIL import Image
        from pyhanko.pdf_utils.images import PdfImage

        img = Image.open(bg_spec)
        # Setting the writer can be delayed
        return PdfImage(img, writer=None)


@dataclass(frozen=True)
class BaseStampStyle(ConfigurableMixin):
    """
    Base class for stamp styles.
    """

    border_width: int = 3
    """
    Border width in user units (for the stamp, not the text box).
    """

    border_color: Optional[Tuple[float, float, float]] = None
    """
    Border color specified as an RGB tuple taking values between 0.0 and 1.0.

    .. warning::
        There is currently no direct support for non-RGB color spaces.
    """

    background: Optional[content.PdfContent] = None
    """
    :class:`~.pdf_utils.content.PdfContent` instance that will be used to render
    the stamp's background.
    """

    background_layout: layout.SimpleBoxLayoutRule = layout.SimpleBoxLayoutRule(
        x_align=layout.AxisAlignment.ALIGN_MID,
        y_align=layout.AxisAlignment.ALIGN_MID,
        margins=layout.Margins.uniform(5),
    )
    """
    Layout rule to render the background inside the stamp's bounding box.
    Only used if the background has a fully specified :attr:`PdfContent.box`.

    Otherwise, the renderer will position the cursor at
    ``(left_margin, bottom_margin)`` and render the content as-is.
    """

    background_opacity: float = 0.6
    """
    Opacity value to render the background at. This should be a floating-point
    number between `0` and `1`.
    """

    @classmethod
    def process_entries(cls, config_dict):
        """
        This implementation of :meth:`process_entries` processes the
        :attr:`background` configuration value.
        This can either be a path to an image file, in which case it will
        be turned into an instance of :class:`~.pdf_utils.images.PdfImage`,
        or the special value ``__stamp__``, which is an alias for
        :const:`~pyhanko.stamp.STAMP_ART_CONTENT`.
        """

        super().process_entries(config_dict)
        bg_spec = None
        try:
            bg_spec = config_dict['background']
        except KeyError:
            pass
        if bg_spec is not None:
            config_dict['background'] = _get_background_content(bg_spec)

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> Optional['BaseStamp']:
        raise NotImplementedError


class BaseStamp(AppearanceContent):
    def __init__(
        self,
        writer: BasePdfFileWriter,
        style,
        box: Optional[layout.BoxConstraints] = None,
    ):
        super().__init__(box=box, writer=writer)
        self.style = style
        self._resources_ready = False

    def _render_background(self):
        bg = self.style.background
        bg.set_writer(self.writer)
        bg_content = bg.render()  # render first, in case the BBox is lazy

        bg_box = bg.box
        if bg_box.width_defined and bg_box.height_defined:
            # apply layout rule
            positioning = self.style.background_layout.fit(
                self.box, bg_box.width, bg_box.height
            )
        else:
            # No idea about the background dimensions, so just use
            # the left/bottom margins and hope for the best
            margins = self.style.background_layout.margins
            positioning = layout.Positioning(
                x_scale=1, y_scale=1, x_pos=margins.left, y_pos=margins.bottom
            )

        # set opacity in graphics state
        opacity = generic.FloatObject(self.style.background_opacity)
        self.set_resource(
            category=content.ResourceType.EXT_G_STATE,
            name=pdf_name('/BackgroundGS'),
            value=generic.DictionaryObject(
                {pdf_name('/CA'): opacity, pdf_name('/ca'): opacity}
            ),
        )

        # Position & render the background
        command = b'q /BackgroundGS gs %s %s Q' % (
            positioning.as_cm(),
            bg_content,
        )
        # we do this after render(), just in case our background resource
        # decides to pull in extra stuff during rendering
        self.import_resources(bg.resources)
        return command

    def _render_inner_content(self):
        raise NotImplementedError

    def render(self):
        command_stream = [b'q']

        inner_content = self._render_inner_content()

        # Now that the inner layout is done, the dimensions of our bounding
        # box should all have been reified. Let's put in the background,
        # if there is one
        if self.style.background:
            command_stream.append(self._render_background())

        # put in the inner content
        if inner_content:
            command_stream.extend(inner_content)

        # draw the border around the stamp
        bbox = self.box
        border_width = self.style.border_width
        border_color = self.style.border_color
        if border_width:
            if border_color:
                command_stream.append(b'%g %g %g RG' % border_color)

            command_stream.append(
                b'%g w 0 0 %g %g re S'
                % (border_width, bbox.width, bbox.height),
            )

        command_stream.append(b'Q')
        return b' '.join(command_stream)

    def apply(
        self,
        dest_page: int,
        x: int,
        y: int,
        *,
        coords: CoordinateSystem = CoordinateSystem.PAGE_DEFAULT,
    ):
        """
        Apply a stamp to a particular page in the PDF writer attached to this
        :class:`.BaseStamp` instance.

        :param dest_page:
            Index of the page to which the stamp is to be applied
            (starting at `0`).
        :param x:
            Horizontal position of the stamp's lower left corner on the page.
        :param y:
            Vertical position of the stamp's lower left corner on the page.
        :param coords:
            The coordinate convention to use.
        :return:
            A reference to the affected page object, together with
            a ``(width, height)`` tuple describing the dimensions of the stamp.
        """
        stamp_ref = self.register()
        resource_name = b'/Stamp' + hexlify(uuid.uuid4().bytes)
        stamp_paint = b'q 1 0 0 1 %g %g cm %s Do Q' % (
            rd(x),
            rd(y),
            resource_name,
        )
        stamp_wrapper_stream = generic.StreamObject(stream_data=stamp_paint)
        resources = generic.DictionaryObject(
            {
                pdf_name('/XObject'): generic.DictionaryObject(
                    {pdf_name(resource_name.decode('ascii')): stamp_ref}
                )
            }
        )
        wr = self.writer
        assert wr is not None
        if coords == CoordinateSystem.PAGE_DEFAULT:
            # wrap the existing content in q / Q
            wr.add_stream_to_page(
                dest_page,
                wr.add_object(
                    generic.StreamObject(stream_data=b"q"),
                ),
                prepend=True,
            )
            wr.add_stream_to_page(
                dest_page,
                wr.add_object(
                    generic.StreamObject(stream_data=b"Q"),
                ),
            )
        page_ref = wr.add_stream_to_page(
            dest_page, wr.add_object(stamp_wrapper_stream), resources
        )
        dims = (self.box.width, self.box.height)
        return page_ref, dims



from asn1crypto import keys, x509
from cryptography.hazmat.primitives import serialization

__all__ = [
    'translate_pyca_cryptography_cert_to_asn1',
    'translate_pyca_cryptography_key_to_asn1',
]


def translate_pyca_cryptography_key_to_asn1(
    private_key,
) -> keys.PrivateKeyInfo:
    # Store the cert and key as generic ASN.1 structures for more
    # "standardised" introspection. This comes at the cost of some encoding/
    # decoding operations, but those should be fairly insignificant in the
    # grand scheme of things.
    #
    # Note: we're not losing any memory protections here:
    #  (https://cryptography.io/en/latest/limitations.html)
    # Arguably, memory safety is nigh impossible to obtain in a Python
    # context anyhow, and people with that kind of Serious (TM) security
    # requirements should be using HSMs to manage keys.
    return keys.PrivateKeyInfo.load(
        private_key.private_bytes(
            serialization.Encoding.DER,
            serialization.PrivateFormat.PKCS8,
            serialization.NoEncryption(),
        )
    )


def translate_pyca_cryptography_cert_to_asn1(cert) -> x509.Certificate:
    return x509.Certificate.load(cert.public_bytes(serialization.Encoding.DER))



from dataclasses import dataclass

from pyhanko.pdf_utils import content, layout
from pyhanko.pdf_utils.writer import BasePdfFileWriter

from .base import BaseStamp, BaseStampStyle

__all__ = ['StaticContentStamp', 'StaticStampStyle']


@dataclass(frozen=True)
class StaticStampStyle(BaseStampStyle):
    """
    Stamp style that does not include any custom parts; it only renders
    the background.
    """

    background_opacity: float = 1.0
    """
    Opacity value to render the background at. This should be a floating-point
    number between `0` and `1`.
    """

    @classmethod
    def from_pdf_file(
        cls, file_name, page_ix=0, **kwargs
    ) -> 'StaticStampStyle':
        """
        Create a :class:`StaticStampStyle` from a page from an external PDF
        document. This is a convenience wrapper around
        :class:`~content.ImportedPdfContent`.

        The remaining keyword arguments are passed to
        :class:`StaticStampStyle`'s init method.

        :param file_name:
            File name of the external PDF document.
        :param page_ix:
            Page index to import. The default is ``0``, i.e. the first page.
        """
        return StaticStampStyle(
            background=content.ImportedPdfPage(file_name, page_ix=page_ix),
            **kwargs,
        )

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'StaticContentStamp':
        return StaticContentStamp(writer=writer, style=self, box=box)


class StaticContentStamp(BaseStamp):
    """Class representing stamps with static content."""

    def __init__(
        self,
        writer: BasePdfFileWriter,
        style: StaticStampStyle,
        box: layout.BoxConstraints,
    ):
        if not (box and box.height_defined and box.width_defined):
            raise layout.LayoutError(
                "StaticContentStamp requires a predetermined bounding box."
            )
        super().__init__(box=box, style=style, writer=writer)

    def _render_inner_content(self):
        return []



from asn1crypto import core, x509
from pyhanko.sign.ades.asn1_util import register_x509_extension

__all__ = [
    'Iso4217CurrencyCode',
    'MonetaryValue',
    'PKIDisclosureStatement',
    'PKIDisclosureStatements',
    'QcCCLegislationCountryCodes',
    'QcCertificateType',
    'QcCertificateTypeId',
    'QcStatement',
    'QcStatementId',
    'QcStatements',
]


# Technically, the qcStatements types aren't AdES-specific,
# but given their status within eIDAS, there's a large overlap in usage.


class QcStatementId(core.ObjectIdentifier):
    _map = {
        # ETSI EN 319 412-5
        '0.4.0.1862.1.1': 'qc_compliance',
        '0.4.0.1862.1.2': 'qc_limit_value',
        '0.4.0.1862.1.3': 'qc_retention_period',
        '0.4.0.1862.1.4': 'qc_sscd',
        '0.4.0.1862.1.5': 'qc_pki_disclosure_statements',
        '0.4.0.1862.1.6': 'qc_type',
        '0.4.0.1862.1.7': 'qc_cc_legislation',
    }


class Iso4217CurrencyCode(core.Choice):
    _alternatives = [
        ('alphabetic', core.PrintableString),
        ('numeric', core.Integer),
    ]


class MonetaryValue(core.Sequence):
    _fields = [
        ('currency', Iso4217CurrencyCode),
        ('amount', core.Integer),
        ('exponent', core.Integer),
    ]


class PKIDisclosureStatement(core.Sequence):
    _fields = [('url', core.IA5String), ('language', core.PrintableString)]


class PKIDisclosureStatements(core.SequenceOf):
    _child_spec = PKIDisclosureStatement


class QcCertificateTypeId(core.ObjectIdentifier):
    _map = {
        '0.4.0.1862.1.6.1': 'qct_esign',
        '0.4.0.1862.1.6.2': 'qct_eseal',
        '0.4.0.1862.1.6.3': 'qct_web',
    }


class QcCertificateType(core.SequenceOf):
    _child_spec = QcCertificateTypeId


class QcCCLegislationCountryCodes(core.SequenceOf):
    _child_spec = core.PrintableString


class QcStatement(core.Sequence):
    _fields = [
        ('statement_id', QcStatementId),
        ('statement_info', core.Any, {'optional': True}),
    ]
    _oid_pair = ('statement_id', 'statement_info')
    _oid_specs = {
        'qc_compliance': core.Void,
        'qc_limit_value': MonetaryValue,
        'qc_retention_period': core.Integer,
        'qc_sscd': core.Void,
        'qc_pki_disclosure_statements': PKIDisclosureStatements,
        'qc_type': QcCertificateType,
        'qc_cc_legislation': QcCCLegislationCountryCodes,
    }


class QcStatements(core.SequenceOf):
    _child_spec = QcStatement


def get_qc_statements(cert: x509.Certificate) -> QcStatements:
    for ext in cert['tbs_certificate']['extensions']:
        if ext['extn_id'].native != 'qc_statements':
            continue
        qc_statements: QcStatements = ext['extn_value'].parsed
        return qc_statements
    else:
        return QcStatements()


register_x509_extension('1.3.6.1.5.5.7.1.3', 'qc_statements', QcStatements)



from dataclasses import dataclass
from io import BytesIO
from typing import List, Tuple

from pyhanko.pdf_utils import misc
from pyhanko.pdf_utils.generic import (
    ContentOpReading,
    OperatorLiteral,
    PdfObject,
    Reference,
    StreamObject,
    read_object,
)
from pyhanko.pdf_utils.misc import skip_over_comments, skip_over_whitespace


@dataclass(frozen=True)
class GraphicsOperator:
    op: str
    args: List[PdfObject]
    start: Tuple[Reference, int]
    end: Tuple[Reference, int]


def parse_content_stream(stream_parts: List[StreamObject]):
    start_ref = None
    start_pos = 0
    args_collected: List[PdfObject] = []
    for pdf_stream in stream_parts:
        current_ref = pdf_stream.container_ref
        assert isinstance(current_ref, Reference)
        stream_data = pdf_stream.data
        stream = BytesIO(stream_data)
        total = len(stream_data)
        skip_over_whitespace(stream, error_on_end_of_stream=False)
        skip_over_comments(stream, error_on_end_of_stream=False)
        while (pos := stream.tell()) < total:
            obj = read_object(
                stream,
                current_ref,
                in_content_stream=ContentOpReading.GRAPHIC_OPS,
            )
            if start_ref is None:
                args_collected = []
                start_ref = current_ref
                start_pos = pos
            if isinstance(obj, OperatorLiteral):
                # TODO try to skip over inline images? Is annoying to do
                #  in pre-2.0 files.
                if obj.literal == 'BI':
                    raise misc.PdfReadError(
                        "Content streams with inline images are not supported"
                    )
                yield GraphicsOperator(
                    obj.literal,
                    args_collected,
                    (start_ref, start_pos),
                    (current_ref, pos + len(obj.literal)),
                )
                start_ref = None
            else:
                args_collected.append(obj)
            skip_over_whitespace(stream, error_on_end_of_stream=False)
            skip_over_comments(stream, error_on_end_of_stream=False)



import enum
from dataclasses import dataclass
from typing import Optional

from pyhanko.sign.validation.qualified.tsp import (
    QcCertType,
    QualifiedServiceInformation,
)

__all__ = [
    'QcPrivateKeyManagementType',
    'QualificationResult',
    'QualifiedStatus',
]


class QcPrivateKeyManagementType(enum.Enum):
    """
    Description of the private key management methodology.
    """

    UNKNOWN = 0
    """
    Private key management methodology unknown/unspecified.
    """

    QSCD = 1
    """
    Declaration indicating that the private key resides in a qualified
    signature creation device (QSCD).
    """

    QSCD_DELEGATED = 2
    """
    Declaration indicating that the private key resides in a QSCD managed
    on behalf of the subject by another party.
    """

    QSCD_BY_POLICY = 3
    """
    QSCD declaration by pre-eIDAS certificate policy.
    """

    @property
    def is_qscd(self) -> bool:
        return self != QcPrivateKeyManagementType.UNKNOWN


@dataclass(frozen=True)
class QualifiedStatus:
    """
    Represents the qualified status of a certificate.
    """

    qualified: bool
    """
    Indicates whether the certificate is to be considered qualified.
    """

    qc_type: QcCertType
    """
    Type of qualified certificate.
    """

    qc_key_security: QcPrivateKeyManagementType
    """
    Indicates whether the CA declares that the private key
    corresponding to this certificate resides in a qualified
    signature creation device (QSCD) or secure signature creation device (SSCD).
    It also indicates whether the QCSD is managed on behalf of the signer,
    if applicable.

    .. warning::
        These terms are functionally interchangeable, the only difference is
        that "SSCD" is pre-eIDAS terminology.
    """


@dataclass(frozen=True)
class QualificationResult:
    """
    Represents the result of a qualification evaluation.
    """

    status: QualifiedStatus
    """
    Status indicator.
    """

    service_definition: Optional[QualifiedServiceInformation]
    """
    Service definition under which the tested object was considered
    qualified.
    """



from .signers import (
    DEFAULT_MD,
    DEFAULT_SIGNING_STAMP_STYLE,
    ExternalSigner,
    PdfSignatureMetadata,
    PdfSigner,
    PdfTimeStamper,
    Signer,
    SimpleSigner,
    async_sign_pdf,
    load_certs_from_pemder,
    sign_pdf,
)

__all__ = [
    'DEFAULT_MD',
    'DEFAULT_SIGNING_STAMP_STYLE',
    'ExternalSigner',
    'PdfSignatureMetadata',
    'PdfSigner',
    'PdfTimeStamper',
    'Signer',
    'SimpleSigner',
    'async_sign_pdf',
    'load_certs_from_pemder',
    'sign_pdf',
]



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Tuple

from .xades import AnyType, ObjectIdentifierType

__NAMESPACE__ = (
    "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#"
)


class CriteriaListTypeassert(Enum):
    ALL = "all"
    AT_LEAST_ONE = "atLeastOne"
    NONE = "none"


class KeyUsageBitTypename(Enum):
    DIGITAL_SIGNATURE = "digitalSignature"
    NON_REPUDIATION = "nonRepudiation"
    KEY_ENCIPHERMENT = "keyEncipherment"
    DATA_ENCIPHERMENT = "dataEncipherment"
    KEY_AGREEMENT = "keyAgreement"
    KEY_CERT_SIGN = "keyCertSign"
    CRL_SIGN = "crlSign"
    ENCIPHER_ONLY = "encipherOnly"
    DECIPHER_ONLY = "decipherOnly"


@dataclass(frozen=True)
class QualifierType:
    uri: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class KeyUsageBitType:
    value: Optional[bool] = field(
        default=None,
        metadata={
            "required": True,
        },
    )
    name: Optional[KeyUsageBitTypename] = field(
        default=None,
        metadata={
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class PoliciesListType:
    policy_identifier: Tuple[ObjectIdentifierType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "PolicyIdentifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class QualifiersType:
    qualifier: Tuple[QualifierType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "Qualifier",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class KeyUsageType:
    key_usage_bit: Tuple[KeyUsageBitType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "KeyUsageBit",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "min_occurs": 1,
            "max_occurs": 9,
        },
    )


@dataclass(frozen=True)
class CriteriaListType:
    key_usage: Tuple[KeyUsageType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "KeyUsage",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
        },
    )
    policy_set: Tuple[PoliciesListType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "PolicySet",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
        },
    )
    criteria_list: Tuple["CriteriaListType", ...] = field(
        default_factory=tuple,
        metadata={
            "name": "CriteriaList",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
        },
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "name": "Description",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
        },
    )
    other_criteria_list: Optional[AnyType] = field(
        default=None,
        metadata={
            "name": "otherCriteriaList",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
        },
    )
    assert_value: Optional[CriteriaListTypeassert] = field(
        default=None,
        metadata={
            "name": "assert",
            "type": "Attribute",
        },
    )


@dataclass(frozen=True)
class QualificationElementType:
    qualifiers: Optional[QualifiersType] = field(
        default=None,
        metadata={
            "name": "Qualifiers",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "required": True,
        },
    )
    criteria_list: Optional[CriteriaListType] = field(
        default=None,
        metadata={
            "name": "CriteriaList",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "required": True,
        },
    )


@dataclass(frozen=True)
class QualificationsType:
    qualification_element: Tuple[QualificationElementType, ...] = field(
        default_factory=tuple,
        metadata={
            "name": "QualificationElement",
            "type": "Element",
            "namespace": "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#",
            "min_occurs": 1,
        },
    )


@dataclass(frozen=True)
class Qualifications(QualificationsType):
    class Meta:
        namespace = "http://uri.etsi.org/TrstSvc/SvcInfoExt/eSigDir-1999-93-EC-TrustedList/#"



import logging
from dataclasses import replace
from typing import Optional, Tuple

from . import generic, layout
from .content import AppearanceContent
from .generic import TextStringObject
from .misc import FormFillingError, PdfReadError
from .rw_common import PdfHandler
from .text import TextBox, TextBoxStyle
from .writer import BasePdfFileWriter

__all__ = [
    'get_single_field_annot',
    'enumerate_fields_in',
    'annot_width_height',
    'find_existing_empty_field',
    'populate_static_text_field',
]

logger = logging.getLogger(__name__)


def get_single_field_annot(
    field: generic.DictionaryObject,
) -> generic.DictionaryObject:
    """
    Internal function to get the annotation of a field.

    :param field:
        A field dictionary.
    :return:
        The dictionary of the corresponding annotation.
    """
    try:
        (annot,) = field['/Kids']
        annot = annot.get_object()
    except (ValueError, TypeError):
        raise FormFillingError(
            "Failed to access form field's annotation. "
            "Form field must have exactly one child annotation, "
            "or it must be combined with its annotation."
        )
    except KeyError:
        annot = field
    return annot


def enumerate_fields_in(
    field_list,
    filled_status=None,
    with_name=None,
    parent_name="",
    parents=None,
    *,
    refs_seen,
    target_field_type,
):
    if not isinstance(field_list, generic.ArrayObject):
        logger.warning(
            f"Values of type {type(field_list)} are not valid as field "
            f"lists, must be array objects -- skipping."
        )
        return

    parents = parents or ()
    for field_ref in field_list:
        if not isinstance(field_ref, generic.IndirectObject):
            logger.warning(
                "Entries in field list must be indirect references -- skipping."
            )
            continue
        if field_ref.reference in refs_seen:
            raise PdfReadError("Circular reference in form tree")

        field = field_ref.get_object()
        if not isinstance(field, generic.DictionaryObject):
            logger.warning(
                "Entries in field list must be dictionary objects, not "
                f"{type(field)} -- skipping."
            )
            continue
        # /T is the field name. If not specified, we're dealing with a bare
        # widget, so skip it. (these should never occur in /Fields, but hey)
        try:
            field_name = field['/T']
        except KeyError:
            continue
        fq_name = (
            field_name
            if not parent_name
            else ("%s.%s" % (parent_name, field_name))
        )
        explicitly_requested = with_name is not None and fq_name == with_name
        child_requested = explicitly_requested or (
            with_name is not None and with_name.startswith(fq_name)
        )
        # /FT is inheritable, so go up the chain
        current_path = (field,) + parents
        for parent_field in current_path:
            try:
                field_type = parent_field['/FT']
                break
            except KeyError:
                continue
        else:
            field_type = None

        if field_type == target_field_type:
            field_value = field.get('/V')
            # "cast" to a regular string object
            filled = bool(field_value)
            status_check = filled_status is None or filled == filled_status
            name_check = with_name is None or explicitly_requested
            if status_check and name_check:
                yield fq_name, field_value, field_ref
        elif explicitly_requested:
            raise FormFillingError(
                f'Field with name {fq_name} exists but is '
                f'not a {target_field_type} field'
            )

        # if necessary, descend into the field hierarchy
        if with_name is None or (child_requested and not explicitly_requested):
            try:
                yield from enumerate_fields_in(
                    field['/Kids'],
                    parent_name=fq_name,
                    parents=current_path,
                    with_name=with_name,
                    filled_status=filled_status,
                    refs_seen=refs_seen | {field_ref.reference},
                    target_field_type=target_field_type,
                )
            except KeyError:
                continue


def annot_width_height(
    annot_dict: generic.DictionaryObject,
) -> Tuple[float, float]:
    """
    Internal function to compute the width and height of an annotation.

    :param annot_dict:
        Annotation dictionary.
    :return:
        a (width, height) tuple
    """
    try:
        x1, y1, x2, y2 = annot_dict['/Rect']
    except KeyError:
        return 0, 0
    w = abs(x1 - x2)
    h = abs(y1 - y2)
    return w, h


DEFAULT_TEXT_FIELD_LAYOUT_RULE = layout.SimpleBoxLayoutRule(
    x_align=layout.AxisAlignment.ALIGN_MIN,
    y_align=layout.AxisAlignment.ALIGN_MIN,
    inner_content_scaling=layout.InnerScaling.NO_SCALING,
)


class TextFieldContent(AppearanceContent):
    def __init__(
        self,
        writer: BasePdfFileWriter,
        text_content: str,
        text_style: TextBoxStyle,
        box: Optional[layout.BoxConstraints] = None,
    ):
        self._text_content = text_content
        box_layout = (
            text_style.box_layout_rule or DEFAULT_TEXT_FIELD_LAYOUT_RULE
        )
        style = replace(
            text_style,
            box_layout_rule=box_layout,
        )
        super().__init__(writer=writer, box=box)

        self.text_box = TextBox(
            style,
            writer=self.writer,
            resources=self.resources,
            box=self.box,
        )

    def render(self) -> bytes:
        self.text_box.content = self._text_content
        rendered = self.text_box.render()
        return rendered


def find_existing_empty_field(
    handler: PdfHandler,
    field_name: str,
    field_type: str,
) -> generic.Reference:
    """
    Find an empty form field of a given type and return a reference to
    the form field dictionary.

    :param handler:
        A PDF handler representing the document to be searched.
    :param field_name:
        The name of the field to look for.
    :param field_type:
        The type of the field to look for.
    :return:
        A reference to the form field dictionary of the resulting field,
        if found.
    :raises FormFillingError:
    if the form field does not exist or is filled.
    """
    try:
        field_list = handler.root['/AcroForm']['/Fields']
    except KeyError:
        raise FormFillingError("No AcroForm present")

    candidates = enumerate_fields_in(
        field_list=field_list,
        with_name=field_name,
        filled_status=False,
        target_field_type=field_type,
        refs_seen=set(),
    )

    try:
        field_name, value, field_ref = next(candidates)
    except StopIteration:
        raise FormFillingError(
            f'No empty text field with name {field_name} found.'
        )

    return field_ref.reference


def populate_static_text_field(
    writer: BasePdfFileWriter,
    field_name: str,
    style: TextBoxStyle,
    content: str,
):
    """
    Populate an existing text field in a PDF document that is to be treated
    as read-only.

    .. warning::

        This function is intended to be used to fill out form fields
        that do not need to be edited interactively later.
        By design, it is not compatible with PDF's variable text mechanisms.

    :param writer:
        PDF document to modify.
    :param field_name:
        The name of the field to look for.
    :param style:
        The text box style to use.
    :param content:
        The content of the text field.
    """
    field_ref = find_existing_empty_field(writer, field_name, field_type='/Tx')
    field = field_ref.get_object()
    assert isinstance(field, generic.DictionaryObject)

    annot_dict = get_single_field_annot(field)

    w, h = annot_width_height(annot_dict)
    content_obj = TextFieldContent(
        writer=writer,
        text_content=content,
        text_style=style,
        box=layout.BoxConstraints(width=w, height=h),
    )
    # TODO option to use existing DR resources and DA?
    content_obj.apply_appearance(annot_dict)
    writer.update_container(annot_dict)

    field['/V'] = TextStringObject(content)

    try:
        flags = int(field['/Ff'])
    except KeyError:
        flags = 0

    field['/Ff'] = generic.NumberObject(flags | 1)
    field.pop('/DA', None)
    writer.mark_update(field_ref)



import enum
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, FrozenSet, Generator, Iterable, Optional, Set, Tuple

from asn1crypto import x509
from pyhanko.sign.validation.settings import KeyUsageConstraints
from pyhanko_certvalidator.authority import (
    Authority,
    AuthorityWithCert,
    TrustAnchor,
    TrustedServiceType,
    TrustQualifiers,
)
from pyhanko_certvalidator.errors import InvalidCertificateError
from pyhanko_certvalidator.path import ValidationPath
from pyhanko_certvalidator.registry import TrustManager

__all__ = [
    'CA_QC_URI',
    'QTST_URI',
    'AdditionalServiceInformation',
    'BaseServiceInformation',
    'CAServiceInformation',
    'CertSubjectDNCriterion',
    'CriteriaCombination',
    'CriteriaList',
    'Criterion',
    'KeyUsageCriterion',
    'PolicySetCriterion',
    'QTSTServiceInformation',
    'QcCertType',
    'Qualification',
    'Qualification',
    'QualifiedServiceInformation',
    'Qualifier',
    'TSPRegistry',
    'TSPServiceParsingError',
    'TSPTrustManager',
]

# noinspection HttpUrlsUsage
_TRSTSVC_URI_BASE = 'http://uri.etsi.org/TrstSvc'
CA_QC_URI = f'{_TRSTSVC_URI_BASE}/Svctype/CA/QC'
QTST_URI = f'{_TRSTSVC_URI_BASE}/Svctype/TSA/QTST'
_TRUSTEDLIST_URI_BASE = f'{_TRSTSVC_URI_BASE}/TrustedList'


class TSPServiceParsingError(ValueError):
    pass


@dataclass(frozen=True)
class AdditionalServiceInformation:
    uri: str
    critical: bool
    textual_info: Optional[str]


class QcCertType(enum.Enum):
    """
    Type of qualified certificate.
    """

    QC_ESIGN = 'qct_esign'
    """
    Certificate qualified for eSignatures.
    """

    QC_ESEAL = 'qct_eseal'
    """
    Certificate qualified for eSeals.
    """

    QC_WEB = 'qct_web'
    """
    Qualified website authentication certificate (QWAC).
    """


_SVCINFOEXT_URI_BASE = f'{_TRUSTEDLIST_URI_BASE}/SvcInfoExt'


@dataclass(frozen=True)
class BaseServiceInformation:
    """
    Common information about a trusted service.
    """

    service_type: str
    """
    The type of service, specified as a URI.

    .. note::
        Corresponds to the ``ServiceTypeIdentifier`` in the trusted list data.
    """

    service_name: str
    """
    Name of the trusted service.
    """

    valid_from: datetime
    """
    Start of the service definition's validity window.
    """

    valid_until: Optional[datetime]
    """
    End of the service definition's validity window,
    if defined. If not, the service is presumed to be
    valid indefinitely.
    """

    provider_certs: Tuple[x509.Certificate, ...]
    """
    Certificates linked to this service provider.
    """

    additional_info_certificate_type: FrozenSet[QcCertType]
    """
    If non-empty, narrows the scope of the specified service type
    to the types of certificate listed.
    """

    other_additional_info: FrozenSet[AdditionalServiceInformation]
    """
    Other information that qualifies the type of service.
    """


class Qualifier(enum.Enum):
    """
    Qualifier as specified in ETSI TS 119 612, 5.5.9.2.
    """

    WITH_SSCD = 'QCWithSSCD'
    NO_SSCD = 'QCNoSSCD'
    SSCD_AS_IN_CERT = 'QCSSCDStatusAsInCert'
    WITH_QSCD = 'QCWithQSCD'
    NO_QSCD = 'QCNoQSCD'
    QSCD_AS_IN_CERT = 'QCQSCDStatusAsInCert'
    QSCD_MANAGED_ON_BEHALF = 'QCQSCDManagedOnBehalf'
    LEGAL_PERSON = 'QCForLegalPerson'
    FOR_ESIG = 'QCForESig'
    FOR_ESEAL = 'QCForESeal'
    FOR_WSA = 'QCForWSA'
    NOT_QUALIFIED = 'NotQualified'
    QC_STATEMENT = 'QCStatement'

    @property
    def uri(self):
        return f"{_SVCINFOEXT_URI_BASE}/{self.value}"


class Criterion:
    """
    Criterion for a qualifier to apply to a certificate.
    """

    def matches(self, cert: x509.Certificate) -> bool:
        """
        Evaluate a certificate against this criterion.

        :param cert:
            The certificate to evaluate.
        :return:
            ``True`` if the criterion matches, ``False`` otherwise.
        """
        raise NotImplementedError


@dataclass(frozen=True)
class KeyUsageCriterion(Criterion):
    """
    Criterion that matches certificates that meet the specified
    key usage constraints.
    """

    settings: KeyUsageConstraints
    """
    Key usage constraint to apply.
    """

    def matches(self, cert: x509.Certificate) -> bool:
        try:
            self.settings.validate(cert)
            return True
        except InvalidCertificateError:
            return False


@dataclass(frozen=True)
class PolicySetCriterion(Criterion):
    """
    Criterion that matches certificates that meet the specified
    certificate policies.
    """

    required_policy_oids: FrozenSet[str]
    """
    Policies that must be applicable to the certificate.

    .. note::
        These OIDs are considered to be specified in the domain
        of the trust root, so they are subject to policy mapping
        in the sense of RFC 5280.
    """

    def matches(self, cert: x509.Certificate) -> bool:
        policy_ext = cert.certificate_policies_value or ()
        found_policies = {pol['policy_identifier'].dotted for pol in policy_ext}
        return self.required_policy_oids.issubset(found_policies)


@dataclass(frozen=True)
class CertSubjectDNCriterion(Criterion):
    required_rdn_part_oids: FrozenSet[str]

    def matches(self, cert: x509.Certificate) -> bool:
        subject_dn: x509.Name = cert.subject
        found_rdn_part_oids = {
            pair['type'].dotted for rdn in subject_dn.chosen for pair in rdn
        }
        return self.required_rdn_part_oids.issubset(found_rdn_part_oids)


@enum.unique
class CriteriaCombination(enum.Enum):
    """
    Defines how to combine sub-criteria.
    """

    ALL = 'all'
    """
    All sub-criteria must match for the criterion to match.
    """

    AT_LEAST_ONE = 'atLeastOne'
    """
    At least one of the sub-criteria must match for the criterion to match.
    """

    NONE = 'none'
    """
    All of the sub-criteria must fail to match for the criterion to match.
    """


@dataclass(frozen=True)
class CriteriaList(Criterion):
    """
    Combine several criteria as one.
    """

    combine_as: CriteriaCombination
    """
    Logical operation to apply to the list of sub-criteria.
    """

    criteria: FrozenSet[Criterion]
    """
    Set of sub-criteria.
    """

    def matches(self, cert: x509.Certificate) -> bool:
        if self.combine_as == CriteriaCombination.ALL:
            return all(c.matches(cert) for c in self.criteria)
        elif self.combine_as == CriteriaCombination.AT_LEAST_ONE:
            return any(c.matches(cert) for c in self.criteria)
        elif self.combine_as == CriteriaCombination.NONE:
            return not any(c.matches(cert) for c in self.criteria)
        else:
            raise NotImplementedError


@dataclass(frozen=True)
class Qualification:
    """
    Representation of a qualification in the sense of ETSI TS 119 612, 5.5.9.2.
    """

    qualifiers: FrozenSet[Qualifier]
    """
    Set of qualifiers to apply to the certificates matching the criteria.
    """

    criteria_list: CriteriaList
    """
    List of criteria to apply.
    """


@dataclass(frozen=True)
class QualifiedServiceInformation:
    """
    Representation of a service with conditional qualifiers.
    """

    base_info: BaseServiceInformation
    """
    Basic information about the service.
    """

    qualifications: FrozenSet[Qualification]
    """
    Relevant qualifications.
    """


@dataclass(frozen=True)
class CAServiceInformation(QualifiedServiceInformation):
    """
    Qualified CA service description.
    """

    # TODO process this setting
    expired_certs_revocation_info: Optional[datetime]
    """
    See ETSI TS 119 612, 5.5.9.1.

    .. warning::
        This extension is not yet taken into account by certificate
        validation processes.
    """


@dataclass(frozen=True)
class QTSTServiceInformation(QualifiedServiceInformation):
    """
    Qualified TSA service description.
    """


def _service_sort_key(si: QualifiedServiceInformation):
    if si.base_info.valid_until is None:
        return 1, None
    else:
        return 0, si.base_info.valid_until


class TSPRegistry:
    """
    Registry of trusted service providers (TSPs), typically populated from
    a trust list.

    Currently, a TSP registry can keep track of qualified CAs and TSAs (QTSTs).
    """

    def __init__(self: 'TSPRegistry'):
        self._ca_cert_to_si: Dict[Authority, Set[CAServiceInformation]] = (
            defaultdict(set)
        )
        self._tst_cert_to_si: Dict[Authority, Set[QTSTServiceInformation]] = (
            defaultdict(set)
        )

    def register_ca(self, ca_service_info: CAServiceInformation):
        """
        Register a trusted certificate authority.

        :param ca_service_info:
            Service information about the CA.
        """
        for cert in ca_service_info.base_info.provider_certs:
            self._ca_cert_to_si[AuthorityWithCert(cert)].add(ca_service_info)

    def register_tst(self, qtst_service_info: QTSTServiceInformation):
        """
        Register a trusted time stamping authority.

        :param qtst_service_info:
            Service information about the TSA.
        """
        for cert in qtst_service_info.base_info.provider_certs:
            self._tst_cert_to_si[AuthorityWithCert(cert)].add(qtst_service_info)

    def applicable_service_definitions(
        self, authority: Authority, moment: Optional[datetime]
    ) -> Iterable[QualifiedServiceInformation]:
        """
        Retrieve the service definitions in this registry on behalf
        of which the specified authority might act.

        .. note::
            This includes all supported types of qualified services,
            not just certificate authorities.

        :param authority:
            Authority to evaluate.
        :param moment:
            Time at which to evaluate the service definitions.
            If none is specified, all matching service definitions
            will be returned, irrespective of their validity windows.
        :return:
            List of service definitions matching the authority.
        """

        all_services = tuple(self._ca_cert_to_si[authority]) + tuple(
            self._tst_cert_to_si[authority]
        )
        sorted_services = sorted(
            all_services, key=_service_sort_key, reverse=True
        )
        for service in sorted_services:
            valid_from = service.base_info.valid_from
            valid_until = service.base_info.valid_until
            if not moment or (
                valid_from <= moment
                and (not valid_until or valid_until >= moment)
            ):
                yield service

    @property
    def known_certificate_authorities(self) -> Iterable[Authority]:
        """
        List known certificate authorities identified by this registry.
        """
        return {ca for ca, sds in self._ca_cert_to_si.items() if sds}

    @property
    def known_timestamp_authorities(self) -> Iterable[Authority]:
        """
        List known time stamping authorities identified by this registry.
        """
        return {tsa for tsa, sds in self._tst_cert_to_si.items() if sds}

    def applicable_tsps_on_path(
        self, path: ValidationPath, moment: datetime
    ) -> Generator[QualifiedServiceInformation, None, None]:
        """
        List applicable trusted service providers on the provided
        validation path.

        :param path:
            The validation path to evaluate.
        :param moment:
            Time at which to evaluate the service definitions.
            If none is specified, all matching service definitions
            will be returned, irrespective of their validity windows.
        :return:
            A generator that yields any service definitions matching the
            authorities on the validation path.
        """
        for ca in path.iter_authorities():
            yield from self.applicable_service_definitions(ca, moment)


class TSPTrustManager(TrustManager):
    """
    Trust manager implementation based on a :class:`.TSPRegistry`.
    """

    def __init__(self, tsp_registry: TSPRegistry):
        self.tsp_registry = tsp_registry

    def as_trust_anchor(self, authority: Authority) -> Optional[TrustAnchor]:
        try:
            # moment = None -> pick the most recent applicable
            sd = next(
                iter(
                    self.tsp_registry.applicable_service_definitions(
                        authority, moment=None
                    )
                )
            )
        except StopIteration:
            return None
        service_type = {
            CA_QC_URI: TrustedServiceType.CERTIFICATE_AUTHORITY,
            QTST_URI: TrustedServiceType.TIME_STAMPING_AUTHORITY,
        }.get(sd.base_info.service_type, TrustedServiceType.UNSUPPORTED)
        return TrustAnchor(
            authority,
            quals=TrustQualifiers(
                trusted_service_type=service_type,
                valid_from=sd.base_info.valid_from,
                valid_until=sd.base_info.valid_until,
            ),
        )

    def find_potential_issuers(
        self, cert: x509.Certificate
    ) -> Generator[TrustAnchor, None, None]:
        for authority in self.tsp_registry.known_certificate_authorities:
            as_anchor = self.as_trust_anchor(authority)
            if as_anchor and authority.is_potential_issuer_of(cert):
                yield as_anchor



import enum

__all__ = ['CoordinateSystem']


class CoordinateSystem(enum.Enum):
    """
    Positioning convention for stamps.
    """

    PAGE_DEFAULT = 1
    """
    Always treat the stamp's position in the page's default coordinate system,
    by defensively forcing a restore to the original graphics state.

    .. note::
        This is the default behaviour since ``0.27.0``.
    """

    AMBIENT = 2
    """
    Apply the stamp in the ambient frame of reference set by the existing
    page content.
    This may yield unpredictable results depending on the input document.

    .. note::
        This was the default behaviour prior to ``0.27.0``.
    """



from dataclasses import dataclass

from pyhanko.pdf_utils import layout
from pyhanko.pdf_utils.writer import BasePdfFileWriter

from .base import BaseStampStyle

__all__ = ['NoOpStampStyle']


@dataclass(frozen=True)
class NoOpStampStyle(BaseStampStyle):
    """
    Stamp style that generates no stamp (preserves existing signature appearance)
    """

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> None:
        return None



import logging
import warnings
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Iterator, Optional, TypeVar

from asn1crypto import cms
from asn1crypto import pdf as asn1_pdf
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko_certvalidator import ValidationContext
from pyhanko_certvalidator.policy_decl import (
    CertRevTrustPolicy,
    RevocationCheckingPolicy,
    RevocationCheckingRule,
)

from ..diff_analysis import DiffPolicy
from ..general import (
    MultivaluedAttributeError,
    NonexistentAttributeError,
    find_unique_cms_attribute,
)
from .dss import DocumentSecurityStore
from .errors import (
    NoDSSFoundError,
    SignatureValidationError,
    ValidationInfoReadingError,
)
from .generic_cms import (
    async_validate_cms_signature,
    cms_basic_validation,
    collect_signer_attr_status,
    validate_tst_signed_data,
)
from .pdf_embedded import EmbeddedPdfSignature, report_seed_value_validation
from .settings import KeyUsageConstraints
from .status import (
    PdfSignatureStatus,
    SignatureStatus,
    TimestampSignatureStatus,
)

__all__ = [
    'RevocationInfoValidationType',
    'apply_adobe_revocation_info',
    'async_validate_pdf_ltv_signature',
    'establish_timestamp_trust',
    'get_timestamp_chain',
    'retrieve_adobe_revocation_info',
]


logger = logging.getLogger(__name__)

StatusType = TypeVar('StatusType', bound=SignatureStatus)


class RevocationInfoValidationType(Enum):
    """
    Indicates a validation profile to use when validating revocation info.
    """

    ADOBE_STYLE = 'adobe'
    """
    Retrieve validation information from the CMS object, using Adobe's
    revocation info archival attribute.
    """

    PADES_LT = 'pades'
    """
    Retrieve validation information from the DSS, and require the signature's
    embedded timestamp to still be valid.
    """

    PADES_LTA = 'pades-lta'
    """
    Retrieve validation information from the DSS, but read & validate the chain
    of document timestamps leading up to the signature to establish the
    integrity of the validation information at the time of signing.
    """

    @classmethod
    def as_tuple(cls):
        return tuple(m.value for m in cls)


DEFAULT_LTV_INTERNAL_REVO_CHECK_POLICY = RevocationCheckingPolicy(
    ee_certificate_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
    intermediate_ca_cert_rule=RevocationCheckingRule.CHECK_IF_DECLARED,
)


def _default_ltv_internal_revo_policy(**kwargs):
    return CertRevTrustPolicy(
        revocation_checking_policy=DEFAULT_LTV_INTERNAL_REVO_CHECK_POLICY,
        **kwargs,
    )


STRICT_LTV_INTERNAL_REVO_CHECK_POLICY = RevocationCheckingPolicy(
    ee_certificate_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
    intermediate_ca_cert_rule=RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
)


def _strict_ltv_internal_revo_policy(**kwargs):
    return CertRevTrustPolicy(
        revocation_checking_policy=STRICT_LTV_INTERNAL_REVO_CHECK_POLICY,
        **kwargs,
    )


def _strict_vc_context_kwargs(
    timestamp: datetime, validation_context_kwargs: dict
):
    # create a new validation context using the timestamp value as the time
    # of evaluation, turn off fetching and load OCSP responses / CRL data
    # from the DSS / revocation info object
    validation_context_kwargs['allow_fetching'] = False
    validation_context_kwargs['moment'] = timestamp

    # Certs with OCSP/CRL endpoints should have the relevant revocation data
    # embedded, if no stricter revocation_mode policy is in place already

    revinfo_policy = validation_context_kwargs.get('revinfo_policy', None)
    retroactive = validation_context_kwargs.get('retroactive_revinfo', False)
    if revinfo_policy is None:
        # handle legacy revocation mode
        legacy_rm = validation_context_kwargs.pop('revocation_mode', None)
        if legacy_rm and legacy_rm != 'soft-fail':
            revinfo_policy = CertRevTrustPolicy(
                RevocationCheckingPolicy.from_legacy(legacy_rm),
            )
        elif legacy_rm == 'soft-fail':
            # fall back to the default
            revinfo_policy = _default_ltv_internal_revo_policy(
                retroactive_revinfo=retroactive
            )
    elif not revinfo_policy.revocation_checking_policy.essential:
        # also in this case, we sub in the default
        revinfo_policy = _strict_ltv_internal_revo_policy(
            retroactive_revinfo=retroactive
        )

    validation_context_kwargs['revinfo_policy'] = revinfo_policy


async def establish_timestamp_trust(
    tst_signed_data: cms.SignedData,
    validation_context: ValidationContext,
    expected_tst_imprint: bytes,
):
    """
    Wrapper around :func:`validate_tst_signed_data` for use when analysing
    timestamps for the purpose of establishing a timestamp chain.
    Its main purpose is throwing/logging an error if validation fails, since
    that amounts to lack of trust in the purported validation time.

    This is internal API.

    :param tst_signed_data:
        The ``SignedData`` value to validate; must encapsulate a ``TSTInfo``
        value.
    :param validation_context:
        The validation context to apply to the timestamp.
    :param expected_tst_imprint:
        The expected message imprint for the ``TSTInfo`` value.
    :return:
        A :class:`.TimestampSignatureStatus` if validation is successful.
    :raises:
        :class:`SignatureValidationError` if validation fails.
    """
    timestamp_status_kwargs = await validate_tst_signed_data(
        tst_signed_data, validation_context, expected_tst_imprint
    )
    timestamp_status = TimestampSignatureStatus(**timestamp_status_kwargs)

    if not timestamp_status.valid or not timestamp_status.trusted:
        logger.warning(
            "Could not validate embedded timestamp token: %s.",
            timestamp_status.summary(),
        )
        raise SignatureValidationError(
            "Could not establish time of signing, timestamp token did not "
            "validate with current settings."
        )
    return timestamp_status


def get_timestamp_chain(
    reader: PdfFileReader,
) -> Iterator[EmbeddedPdfSignature]:
    """
    Get the document timestamp chain of the associated reader, ordered
    from new to old.

    :param reader:
        A :class:`.PdfFileReader`.
    :return:
        An iterable of
        :class:`~pyhanko.sign.validation.pdf_embedded.EmbeddedPdfSignature`
        objects representing document timestamps.
    """
    return filter(
        lambda sig: sig.sig_object.get('/Type', None) == '/DocTimeStamp',
        reversed(reader.embedded_signatures),
    )


@dataclass
class _TimestampTrustData:
    latest_dts: Optional[EmbeddedPdfSignature]
    earliest_ts_status: Optional[TimestampSignatureStatus]
    ts_chain_length: int
    current_signature_vc_kwargs: dict


def _instantiate_ltv_vc(
    emb_timestamp: EmbeddedPdfSignature, validation_context_kwargs
):
    try:
        hist_resolver = emb_timestamp.reader.get_historical_resolver(
            emb_timestamp.signed_revision
        )
        dss = DocumentSecurityStore.read_dss(hist_resolver)
        return dss.as_validation_context(validation_context_kwargs)
    except NoDSSFoundError:
        validation_context_kwargs.setdefault('crls', ())
        validation_context_kwargs.setdefault('ocsps', ())
        return ValidationContext(**validation_context_kwargs)


async def _establish_timestamp_trust_lta(
    reader,
    bootstrap_validation_context,
    validation_context_kwargs,
    until_revision,
) -> _TimestampTrustData:
    timestamps = get_timestamp_chain(reader)
    validation_context_kwargs = dict(validation_context_kwargs)
    current_vc = bootstrap_validation_context
    ts_status = None
    ts_count = -1
    emb_timestamp = None
    for ts_count, emb_timestamp in enumerate(timestamps):
        if emb_timestamp.signed_revision < until_revision:
            break

        external_digest = emb_timestamp.compute_digest()
        ts_status = await establish_timestamp_trust(
            emb_timestamp.signed_data, current_vc, external_digest
        )
        # set up the validation kwargs for the next iteration
        _strict_vc_context_kwargs(
            ts_status.timestamp, validation_context_kwargs
        )
        # read the DSS at the current revision into a new
        # validation context object
        current_vc = _instantiate_ltv_vc(
            emb_timestamp, validation_context_kwargs
        )

    return _TimestampTrustData(
        latest_dts=emb_timestamp,
        earliest_ts_status=ts_status,
        ts_chain_length=ts_count + 1,
        current_signature_vc_kwargs=validation_context_kwargs,
    )


async def async_validate_pdf_ltv_signature(
    embedded_sig: EmbeddedPdfSignature,
    validation_type: RevocationInfoValidationType,
    validation_context_kwargs: Optional[dict] = None,
    bootstrap_validation_context: Optional[ValidationContext] = None,
    ac_validation_context_kwargs=None,
    force_revinfo=False,
    diff_policy: Optional[DiffPolicy] = None,
    key_usage_settings: Optional[KeyUsageConstraints] = None,
    skip_diff: bool = False,
) -> PdfSignatureStatus:
    """
    .. deprecated:: 0.31.0
        Use :func:`~pyhanko.sign.validation.ades.ades_lta_validation` instead.

    .. versionadded:: 0.9.0

    Validate a PDF LTV signature according to a particular profile.

    :param embedded_sig:
        Embedded signature to evaluate.
    :param validation_type:
        Validation profile to use.
    :param validation_context_kwargs:
        Keyword args to instantiate
        :class:`.pyhanko_certvalidator.ValidationContext` objects needed over
        the course of the validation.
    :param ac_validation_context_kwargs:
        Keyword arguments for the validation context to use to
        validate attribute certificates.
        If not supplied, no AC validation will be performed.

        .. note::
            :rfc:`5755` requires attribute authority trust roots to be specified
            explicitly; hence why there's no default.
    :param bootstrap_validation_context:
        Validation context used to validate the current timestamp.
    :param force_revinfo:
        Require all certificates encountered to have some form of live
        revocation checking provisions.
    :param diff_policy:
        Policy to evaluate potential incremental updates that were appended
        to the signed revision of the document.
        Defaults to
        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
    :param key_usage_settings:
        A :class:`.KeyUsageConstraints` object specifying which key usages
        must or must not be present in the signer's certificate.
    :param skip_diff:
        If ``True``, skip the difference analysis step entirely.
    :return:
        The status of the signature.
    """

    warnings.warn(
        "'async_validate_pdf_ltv_signature' uses a deprecated "
        "ad-hoc validation methodology, use 'ades_lta_validation' instead "
        "for a more standards-based approach. This function may be removed "
        "in a future pyHanko version.",
        DeprecationWarning,
    )

    # create a fresh copy of the validation_kwargs
    vc_kwargs: dict = dict(validation_context_kwargs or {})

    # To validate the first timestamp, allow fetching by default
    # we'll turn it off later
    vc_kwargs.setdefault('allow_fetching', True)
    vc_kwargs.setdefault('retroactive_revinfo', False)
    # same for revocation_mode: if force_revinfo is false, we simply turn on
    # hard-fail by default for now. Once the timestamp is validated,
    # we switch to hard-fail forcibly.
    retroactive = vc_kwargs['retroactive_revinfo']
    if force_revinfo:
        vc_kwargs['revinfo_policy'] = _default_ltv_internal_revo_policy(
            retroactive_revinfo=retroactive
        )
        if ac_validation_context_kwargs is not None:
            ac_validation_context_kwargs['revinfo_policy'] = (
                _strict_ltv_internal_revo_policy(
                    retroactive_revinfo=retroactive
                )
            )
    elif 'revocation_mode' not in vc_kwargs:
        vc_kwargs.setdefault(
            'revinfo_policy',
            _default_ltv_internal_revo_policy(retroactive_revinfo=retroactive),
        )
        if ac_validation_context_kwargs is not None:
            ac_validation_context_kwargs.setdefault(
                'revinfo_policy',
                _default_ltv_internal_revo_policy(
                    retroactive_revinfo=retroactive
                ),
            )

    current_vc: ValidationContext
    reader = embedded_sig.reader
    if validation_type == RevocationInfoValidationType.ADOBE_STYLE:
        dss = None
        current_vc = bootstrap_validation_context or ValidationContext(
            **vc_kwargs
        )
    else:
        # If there's a DSS, there's no harm in reading additional certs from it
        dss = DocumentSecurityStore.read_dss(reader)
        if bootstrap_validation_context is None:
            current_vc = dss.as_validation_context(
                vc_kwargs, include_revinfo=False
            )
        else:
            current_vc = bootstrap_validation_context
            # add the certs from the DSS
            current_vc.certificate_registry.register_multiple(dss.load_certs())

    embedded_sig.compute_digest()
    embedded_sig.compute_tst_digest()

    # If the validation profile is PAdES-type, then we validate the timestamp
    #  chain now.
    #  This is bootstrapped using the current validation context.
    #  If successful, we obtain a new validation context set to a new
    #  "known good" verification time. We then repeat the process using this
    #  new validation context instead of the current one.
    # also record the embedded sig object assoc. with the oldest applicable
    # DTS in the timestamp chain
    earliest_good_timestamp_st: Optional[TimestampSignatureStatus] = None
    latest_dts: Optional[EmbeddedPdfSignature] = None
    ts_chain_length = 0
    if validation_type != RevocationInfoValidationType.ADOBE_STYLE:
        ts_trust_data = await _establish_timestamp_trust_lta(
            reader,
            current_vc,
            validation_context_kwargs=vc_kwargs,
            until_revision=embedded_sig.signed_revision,
        )
        ts_chain_length = ts_trust_data.ts_chain_length
        vc_kwargs = ts_trust_data.current_signature_vc_kwargs
        # if we found a timestamp, use it to pass to the next VC iteration
        if ts_trust_data.latest_dts:
            current_vc = _instantiate_ltv_vc(
                ts_trust_data.latest_dts, vc_kwargs
            )
        # In PAdES-LTA, we should only rely on DSS information that is covered
        # by an appropriate document timestamp.
        # If the validation profile is PAdES-LTA, then we must have seen
        # at least one document timestamp pass by, i.e. earliest_known_timestamp
        # must be non-None by now.
        if (
            ts_trust_data.earliest_ts_status is None
            and validation_type == RevocationInfoValidationType.PADES_LTA
        ):
            raise SignatureValidationError(
                "Purported PAdES-LTA signature does not have a timestamp chain."
            )
        # if this assertion fails, there's a bug in the validation code
        assert (
            validation_type == RevocationInfoValidationType.PADES_LT
            or ts_trust_data.ts_chain_length >= 1
        )
        earliest_good_timestamp_st = ts_trust_data.earliest_ts_status
        latest_dts = ts_trust_data.latest_dts

    # now that we have arrived at the revision with the signature,
    # we can check for a timestamp token attribute there
    # (This is allowed, regardless of whether we use Adobe-style LTV or
    # a PAdES validation profile)
    tst_signed_data = embedded_sig.attached_timestamp_data
    if tst_signed_data is not None:
        tst_signature_digest = embedded_sig.tst_signature_digest
        assert tst_signature_digest is not None
        earliest_good_timestamp_st = await establish_timestamp_trust(
            tst_signed_data, current_vc, tst_signature_digest
        )
        assert earliest_good_timestamp_st is not None
        # update validation context moments to the new signature POE
        signature_poe = earliest_good_timestamp_st.timestamp
        vc_kwargs['moment'] = signature_poe
        if ac_validation_context_kwargs is not None:
            ac_validation_context_kwargs['moment'] = signature_poe
    elif (
        validation_type == RevocationInfoValidationType.PADES_LTA
        and ts_chain_length == 1
    ):
        # TODO Pretty sure that this is the spirit of the LTA profile,
        #  but are we being too harsh here? I don't think so, but it's worth
        #  revisiting later
        # For later review: I believe that this check is appropriate, because
        # the timestamp that protects the signature should be verifiable
        # using only information from the next DSS, which should in turn
        # also be protected using a DTS. This requires at least two timestamps.
        raise SignatureValidationError(
            "PAdES-LTA signature requires separate timestamps protecting "
            "the signature & the rest of the revocation info."
        )

    # if, by now, we still don't have a trusted timestamp, there's a problem
    # regardless of the validation profile in use.
    if earliest_good_timestamp_st is None:
        raise SignatureValidationError(
            'LTV signatures require a trusted timestamp.'
        )

    _strict_vc_context_kwargs(earliest_good_timestamp_st.timestamp, vc_kwargs)

    stored_ac_vc = None
    if validation_type == RevocationInfoValidationType.ADOBE_STYLE:
        ocsps, crls = retrieve_adobe_revocation_info(embedded_sig.signer_info)
        vc_kwargs['ocsps'] = ocsps
        vc_kwargs['crls'] = crls
        stored_vc = ValidationContext(**vc_kwargs)
        if ac_validation_context_kwargs is not None:
            ac_validation_context_kwargs['ocsps'] = ocsps
            ac_validation_context_kwargs['crls'] = crls
            stored_ac_vc = ValidationContext(**ac_validation_context_kwargs)
    elif validation_type == RevocationInfoValidationType.PADES_LT:
        # in this case, we don't care about whether the information
        # in the DSS is protected by any timestamps, so just ingest everything
        assert dss is not None
        stored_vc = dss.as_validation_context(vc_kwargs)
        if ac_validation_context_kwargs is not None:
            stored_ac_vc = dss.as_validation_context(
                ac_validation_context_kwargs
            )
    else:
        # in the LTA profile, we should use only DSS information covered
        # by the last relevant timestamp, so the correct VC is encoded
        # by the current validation context kwargs retrieved by the lta helper
        # logic (note that we updated the POE timestamps further up)
        assert latest_dts is not None
        stored_vc = _instantiate_ltv_vc(latest_dts, vc_kwargs)
        if ac_validation_context_kwargs is not None:
            stored_ac_vc = _instantiate_ltv_vc(
                latest_dts, ac_validation_context_kwargs
            )

    # Now, we evaluate the validity of the timestamp guaranteeing the signature
    #  *within* the LTV context.
    #   (i.e. we check whether there's enough revinfo to keep tabs on the
    #   timestamp's validity)
    # If the last timestamp comes from a timestamp token attached to the
    # signature, it should be possible to validate it using only data from the
    # DSS / revocation info store, so validate the timestamp *again*
    # using those settings.

    if (
        tst_signed_data is not None
        or validation_type == RevocationInfoValidationType.PADES_LT
    ):
        if tst_signed_data is not None:
            ts_to_validate = tst_signed_data
        else:
            # we're in the PAdES-LT case with a detached TST now.
            # this should be conceptually equivalent to the above
            # so we run the same check here
            assert latest_dts is not None
            ts_to_validate = latest_dts.signed_data
        ts_status_coro = async_validate_cms_signature(
            ts_to_validate,
            status_cls=TimestampSignatureStatus,
            validation_context=stored_vc,
            status_kwargs={'timestamp': earliest_good_timestamp_st.timestamp},
        )
        timestamp_status: TimestampSignatureStatus = await ts_status_coro
    else:
        # In the LTA case, we don't have to do any further checks, since the
        # _establish_timestamp_trust_lta handled that for us.
        # We can therefore just take earliest_good_timestamp_st at face value.
        timestamp_status = earliest_good_timestamp_st

    embedded_sig.compute_integrity_info(
        diff_policy=diff_policy, skip_diff=skip_diff
    )
    status_kwargs = embedded_sig.summarise_integrity_info()
    status_kwargs.update(
        {
            'signer_reported_dt': earliest_good_timestamp_st.timestamp,
            'timestamp_validity': timestamp_status,
        }
    )
    key_usage_settings = PdfSignatureStatus.default_usage_constraints(
        key_usage_settings
    )
    status_kwargs = await cms_basic_validation(
        embedded_sig.signed_data,
        raw_digest=embedded_sig.external_digest,
        validation_context=stored_vc,
        status_kwargs=status_kwargs,
        key_usage_settings=key_usage_settings,
    )

    report_seed_value_validation(
        embedded_sig, status_kwargs['validation_path'], timestamp_found=True
    )
    if stored_ac_vc is not None:
        stored_ac_vc.certificate_registry.register_multiple(
            embedded_sig.other_embedded_certs
        )
    status_kwargs.update(
        await collect_signer_attr_status(
            sd_attr_certificates=embedded_sig.embedded_attr_certs,
            signer_cert=embedded_sig.signer_cert,
            validation_context=stored_ac_vc,
            sd_signed_attrs=embedded_sig.signer_info['signed_attrs'],
        )
    )

    return PdfSignatureStatus(**status_kwargs)


def retrieve_adobe_revocation_info(signer_info: cms.SignerInfo):
    """
    Retrieve Adobe-style revocation information from a ``SignerInfo`` value,
    if present.

    Internal API.

    :param signer_info:
        A ``SignerInfo`` value.
    :return:
        A tuple of two (potentially empty) lists, containing OCSP
        responses and CRLs, respectively.
    """
    try:
        revinfo: asn1_pdf.RevocationInfoArchival = find_unique_cms_attribute(
            signer_info['signed_attrs'], "adobe_revocation_info_archival"
        )
    except (NonexistentAttributeError, MultivaluedAttributeError) as e:
        raise ValidationInfoReadingError(
            "No revocation info archival attribute found, or multiple present"
        ) from e

    ocsps = list(revinfo['ocsp'] or ())
    crls = list(revinfo['crl'] or ())
    return ocsps, crls


def apply_adobe_revocation_info(
    signer_info: cms.SignerInfo, validation_context_kwargs=None
) -> ValidationContext:
    """
    Read Adobe-style revocation information from a CMS object, and load it
    into a validation context.

    :param signer_info:
        Signer info CMS object.
    :param validation_context_kwargs:
        Extra kwargs to pass to the ``__init__`` function.
    :return:
        A validation context preloaded with the relevant revocation information.
    """
    validation_context_kwargs = validation_context_kwargs or {}
    ocsps, crls = retrieve_adobe_revocation_info(signer_info)
    return ValidationContext(
        ocsps=ocsps, crls=crls, **validation_context_kwargs
    )
